from .data_utils import ensure_data_downloaded
ensure_data_downloaded()

import os
datadir = os.path.join(os.path.dirname(__file__), "data")

from . import _ssapy

from .orbit import Orbit, EarthObserver, OrbitalObserver
from .propagator import (
    KeplerianPropagator, SeriesPropagator, RK4Propagator, SGP4Propagator,
    SciPyPropagator, RK78Propagator, RK8Propagator
)
from .compute import rv, dircos, radec, altaz, quickAltAz, radecRate, groundTrack

from .accel import Accel, AccelKepler, AccelSum, AccelEarthRad, AccelSolRad, AccelDrag, AccelConstNTW

from .linker import ModelSelectorParams, BinarySelectorParams, Linker

from .orbit_solver import TwoPosOrbitSolver, GaussTwoPosOrbitSolver
from .orbit_solver import DanchickTwoPosOrbitSolver, SheferTwoPosOrbitSolver
from .orbit_solver import ThreeAngleOrbitSolver

from .particles import Particles

from .rvsampler import GEOProjectionInitializer, DistanceProjectionInitializer
from .rvsampler import circular_guess
from .rvsampler import GaussianRVInitializer, DirectInitializer
from .rvsampler import RVProbability, EmceeSampler, MVNormalProposal
from .rvsampler import RVSigmaProposal, MHSampler, LMOptimizer

from .ellipsoid import Ellipsoid

from .body import (
    EarthOrientation, MoonOrientation, MoonPosition, Body, get_body
)
from .gravity import HarmonicCoefficients, AccelThirdBody, AccelHarmonic

# Some items we keep out of the ssa namespace, but still provide through
# submodules
from . import constants
from . import plotUtils
from . import io
from . import utils
from . import simple
from astropy.time import Time, TimeDelta
import astropy.units as u
from datetime import timedelta





import ssapy
import os
import shutil
import numpy as np
import matplotlib.pyplot as plt

save_folder = './ssapy_test_plots'
print(f"Putting test_plot.py output in: {save_folder}")

temp_directory = f'{save_folder}/rotate_vector_frames/'

# Check if the directory exists, and create it if it doesn't
try:
    if not os.path.exists(temp_directory):
        os.makedirs(temp_directory)
        print(f"Created directory: {temp_directory}")
    else:
        print(f"Directory already exists: {temp_directory}")
except Exception as e:
    print(f"Error creating directory: {e}")

# Testing rotate_vector() in utils.
v_unit = np.array([1, 0, 0])  # Replace this with your actual unit vector

figs = []

i = 0
for theta in range(0, 181, 20):
    for phi in range(0, 361, 20):
        new_unit_vector = ssapy.utils.rotate_vector(v_unit, theta, phi, plot_path=temp_directory, save_idx=i)
        i += 1

gif_path = f"{save_folder}/rotate_vectors_{v_unit[0]:.0f}_{v_unit[1]:.0f}_{v_unit[2]:.0f}.gif"
ssapy.plotUtils.save_animated_gif(gif_name=gif_path, frames=ssapy.io.listdir(f'{temp_directory}*', sorted=True), fps=20)
# shutil.rmtree(temp_directory)


# Creating orbit plots
times = ssapy.utils.get_times(duration=(1, 'year'), freq=(1, 'hour'), t0='2025-3-1')
moon = ssapy.get_body("moon").position(times).T


def initialize_DRO(t, delta_r=7.52064e7, delta_v=344):
    """
    Calculate a distant retrograde orbit (DRO) as an orbit with
    adjustments based on the Moon's position and velocity.

    Parameters:
    ----------
    t : Time
        The time at which to calculate the orbit.
    delta_r : float, optional
        The adjustment to the Moon's position (default is 7.52064e7 meters).
    delta_v : float, optional
        The adjustment to the Moon's velocity (default is 344 meters/second).

    Returns:
    -------
    Orbit
        SSAPy Orbit object.
    """
    moon = ssapy.get_body("moon")

    unit_vector_moon = moon.position(t) / np.linalg.norm(moon.position(t))
    moon_v = (moon.position(t.gps) - moon.position(t.gps - 1)) / 1
    unit_vector_moon_velocity = moon_v / np.linalg.norm(moon_v)
    ssapy.compute.lunar_lagrange_points(t=times[0])

    r = (np.linalg.norm(moon.position(t)) - delta_r) * unit_vector_moon
    v = (np.linalg.norm(moon_v) + delta_v) * unit_vector_moon_velocity

    orbit = ssapy.Orbit(r=r, v=v, t=t)
    return orbit


# Distant Retrograde Orbit (DRO)
dro_orbit = initialize_DRO(t=times[0])
r, v, t = ssapy.simple.ssapy_orbit(orbit=dro_orbit, t=times)
ssapy.plotUtils.orbit_plot(r=r, t=times, save_path=f"{save_folder}/DRO_orbit", frame='Lunar', show=False)
r_lunar, v_lunar = ssapy.utils.gcrf_to_lunar_fixed(r, t=times, v=True)
print("Successfully converted GCRF to lunar frame.")
ssapy.plotUtils.koe_plot(r, v, t=times, body='Earth', save_path=f"{save_folder}Keplerian_orbital_elements.png")

ssapy.plotUtils.orbit_plot(r=r, t=times, save_path=f"{save_folder}/gcrf_plot.png", frame='gcrf', show=True)
ssapy.plotUtils.orbit_plot(r=r, t=times, save_path=f"{save_folder}/itrf_plot", frame='itrf', show=True)
ssapy.plotUtils.orbit_plot(r=r, t=times, save_path=f"{save_folder}/lunar_plot", frame='lunar', show=True)
ssapy.plotUtils.orbit_plot(r=r, t=times, save_path=f"{save_folder}/lunar_axis_lot", frame='lunar axis', show=True)
print("Created a GCRF orbit plot.")
print("Created a ITRF orbit plot.")
print("Created a Lunar orbit plot.")
print("Created a Lunar axis orbit plot.")

# Globe plot of a Geostationary Transfer Orbit (GTO)
r_geo, _, t_geo = ssapy.simple.ssapy_orbit(a=ssapy.constants.RGEO, e=0.3, t=times)
ssapy.plotUtils.globe_plot(r=r_geo, t=t_geo, save_path=f"{save_folder}/globe_plot", scale=5)
print('Created a globe plot.')

ssapy.plotUtils.ground_track_plot(r=r_geo, t=t_geo, ground_stations=None, save_path=f"{save_folder}/ground_track_plot")
print('Created a ground track plot.')

# Example usage
earth_pos = np.array([0, 0, 0])  # Earth at the origin
moon_pos = ssapy.get_body("moon").position(times[0]).T

# Plotting
fig = plt.figure(figsize=(8, 8))
fig.patch.set_facecolor('white')
ax = fig.add_subplot(111, projection='3d')

# Plot Earth
ax.scatter(earth_pos[0], earth_pos[1], earth_pos[2], color='blue', label='Earth')
ax.text(earth_pos[0], earth_pos[1], earth_pos[2], 'Earth', color='blue')

# Plot Moon
ax.scatter(moon_pos[0], moon_pos[1], moon_pos[2], color='grey', label='Moon')
ax.text(moon_pos[0], moon_pos[1], moon_pos[2], 'Moon', color='grey')

# Plot Lagrange points
colors = ['red', 'green', 'purple', 'orange', 'cyan']
for (point, pos), color in zip(ssapy.compute.lunar_lagrange_points(t=times[0]).items(), colors):
    ax.scatter(pos[0], pos[1], pos[2], color=color, label=point)
    ax.text(pos[0], pos[1], pos[2], point, color=color)

# Add a dashed black circle at the lunar distance (LD)
current_LD = np.linalg.norm(moon_pos, axis=-1)
normal_vector = ssapy.compute.moon_normal_vector(t=times[0])
ssapy.plotUtils.draw_dashed_circle(ax, normal_vector, current_LD, 12)
ax.quiver(0, 0, 0, normal_vector[0], normal_vector[1], normal_vector[2], color='r', length=1)

# Labels and legend
ax.set_xlabel('X (m)')
ax.set_ylabel('Y (m)')
ax.set_zlabel('Z (m)')
ax.set_title("Lunar Lagrange Points using Moon's true position")
ax.axis('equal')
ax.legend()
plt.show()
ssapy.plotUtils.save_plot(fig, save_path=f"{save_folder}/lagrange_points")

# Plotting
fig = plt.figure(figsize=(8, 8))
fig.patch.set_facecolor('white')
ax = fig.add_subplot(111, projection='3d')

# Plot Earth
ax.scatter(earth_pos[0], earth_pos[1], earth_pos[2], color='blue', label='Earth')
ax.text(earth_pos[0], earth_pos[1], earth_pos[2], 'Earth', color='blue')

# Plot Moon
ax.scatter(moon_pos[0], moon_pos[1], moon_pos[2], color='grey', label='Moon')
ax.text(moon_pos[0], moon_pos[1], moon_pos[2], 'Moon', color='grey')

# Plot Lagrange points
colors = ['red', 'green', 'purple', 'orange', 'cyan']
for (point, pos), color in zip(ssapy.compute.lunar_lagrange_points_circular(t=times[0]).items(), colors):
    ax.scatter(pos[0], pos[1], pos[2], color=color, label=point)
    ax.text(pos[0], pos[1], pos[2], point, color=color)

# Add a dashed black circle at distance LD
current_LD = np.linalg.norm(moon_pos, axis=-1)
normal_vector = ssapy.compute.moon_normal_vector(t=times[0])
ssapy.plotUtils.draw_dashed_circle(ax, normal_vector, current_LD, 12)
ax.quiver(0, 0, 0, normal_vector[0], normal_vector[1], normal_vector[2], color='r', length=1)

# Labels and legend
ax.set_xlabel('X (m)')
ax.set_ylabel('Y (m)')
ax.set_zlabel('Z (m)')
ax.set_title('Lunar Lagrange Points assuming circular orbit.')
ax.axis('equal')
ax.legend()
plt.show()
ssapy.plotUtils.save_plot(fig, save_path=f"{save_folder}/lagrange_points")

print("Lagrange points were calculated correctly.")
print("Rotate vector plot successfully created.")
print("save_plot() executed successfully.")
print("save_animated_gif() executed successfully.")

print("\nFinished plot testing!\n")



"""
Class to handle transformations between ECEF x,y,z coords and geodetic
longitude, latitude, and height.

Technically, only handles a one-axis ellipsoid, defined via a flattening
parameter f, but that's good enough for simple Earth models.
"""

import numpy as np
from .utils import continueClass
from . import _ssapy
from ._ssapy import Ellipsoid   


@continueClass
class Ellipsoid:
    """
    A class representing an ellipsoid, providing methods to convert between spherical and Cartesian coordinates.

    Methods
    -------
    sphereToCart(lon, lat, height)
        Converts spherical coordinates (longitude, latitude, height) to Cartesian coordinates (x, y, z).
    
    cartToSphere(x, y, z)
        Converts Cartesian coordinates (x, y, z) to spherical coordinates (longitude, latitude, height).
    """
    
    def sphereToCart(self, lon, lat, height):
        """
        Converts spherical coordinates to Cartesian coordinates.

        Parameters
        ----------
        lon : array-like
            Longitude values in radians.
        lat : array-like
            Latitude values in radians.
        height : array-like
            Height above the ellipsoid surface.

        Returns
        -------
        tuple of array-like
            A tuple containing:
            - x : array-like
                Cartesian x-coordinate.
            - y : array-like
                Cartesian y-coordinate.
            - z : array-like
                Cartesian z-coordinate.

        Notes
        -----
        This method uses broadcasting to handle inputs of varying shapes and ensures 
        contiguous arrays for efficient computation.
        """

        lon, lat, height = np.broadcast_arrays(lon, lat, height)
        lon = np.ascontiguousarray(lon)
        lat = np.ascontiguousarray(lat)
        height = np.ascontiguousarray(height)

        x = np.empty_like(lon)
        y = np.empty_like(lon)
        z = np.empty_like(lon)

        self._sphereToCart(
            lon.ctypes.data, lat.ctypes.data, height.ctypes.data, lon.size,
            x.ctypes.data, y.ctypes.data, z.ctypes.data
        )

        return x, y, z

    def cartToSphere(self, x, y, z):
        """
        Converts Cartesian coordinates to spherical coordinates.

        Parameters
        ----------
        x : array-like
            Cartesian x-coordinate.
        y : array-like
            Cartesian y-coordinate.
        z : array-like
            Cartesian z-coordinate.

        Returns
        -------
        tuple of array-like
            A tuple containing:
            - lon : array-like
                Longitude values in radians.
            - lat : array-like
                Latitude values in radians.
            - height : array-like
                Height above the ellipsoid surface.

        Notes
        -----
        This method uses broadcasting to handle inputs of varying shapes and ensures 
        contiguous arrays for efficient computation.
        """

        x, y, z = np.broadcast_arrays(x, y, z)
        x = np.ascontiguousarray(x)
        y = np.ascontiguousarray(y)
        z = np.ascontiguousarray(z)

        lon = np.empty_like(x)
        lat = np.empty_like(x)
        height = np.empty_like(x)

        self._cartToSphere(
            x.ctypes.data, y.ctypes.data, z.ctypes.data, x.size,
            lon.ctypes.data, lat.ctypes.data, height.ctypes.data
        )

        return lon, lat, height






"""
Classes for propagating orbits.
"""

from abc import ABC, abstractmethod
import numpy as np
from astropy.time import Time

from .utils import LRU_Cache, norm, teme_to_gcrf


# Set up a cache for propagation interpolants.  Useful in case one wants to
# propagate using the same Propagator and Orbit multiple times, such as when
# optimizing a position along an orbit to find a rise-time, set-time,
# highest-elevation, etc., or when computing topocentric quantities for one
# orbit and many observers.

# The cache will be a map (orbit, propagator) -> interpolant.  One slight
# wrinkle is that we need to be able to extend the domain of a previously cached
# interpolant.  An easy way to implement this is to have the return object be a
# mutable container (such as a python list) instead of the raw interpolant.  We
# can then mutate the interpolant without invalidating the cached container.


def _getInterpolantContainer(orbit, propagator):
    # The default contained interpolant is empty.  Calling code should mutate
    # the container to hold a useful interpolant.  Because the list is mutated
    # in-place, it remains in the cache in the same position.
    return []


_InterpolantCache = LRU_Cache(_getInterpolantContainer)


class Propagator(ABC):
    """ Abstract base class for orbit propagators.

    Note, the interface for propagators is through functions in ssapy.compute.
    """
    # Essentially no public interface.
    # Interface is instead through compute.rv

    @abstractmethod
    def _getRVOne(self, orbit, time):
        ...  # Subclasses must override

    def _getRVMany(self, orbit, time):
        # Computes positions, velocities broadcasting both over orbits and
        # times.  Base class version is just a for-loop, but subclasses may
        # override if they can be more efficient (e.g., KeplerianPropagator).
        nOrbit = len(orbit)
        nTime = len(time)
        outR = np.empty((nOrbit, nTime, 3,), dtype=float)
        outV = np.empty_like(outR)
        for j, orb in enumerate(orbit):
            outR[j], outV[j] = self._getRVOne(orb, time)
        return outR, outV


class KeplerianPropagator(Propagator):
    """ A basic Keplerian propagator for finding the position and velocity of an
    orbiting object at some future or past time.
    """
    def __repr__(self):
        return "KeplerianPropagator()"

    def _getRVOne(self, orbit, time):
        from .orbit import _ellipticalMeanToEccentricLongitude
        from .orbit import _hyperbolicMeanToEccentricLongitude
        dts = np.atleast_1d(time - orbit.t)
        lM = orbit.lM + orbit.meanMotion * dts
        if orbit.a > 0:
            lE = _ellipticalMeanToEccentricLongitude(lM, orbit.ex, orbit.ey)
        else:
            lE = _hyperbolicMeanToEccentricLongitude(lM, orbit.ex, orbit.ey)
        r, v = orbit._rvFromEquinoctial(lE=np.atleast_2d(lE))
        return r[0], v[0]

    def _getRVMany(self, orbit, time):
        from .orbit import _ellipticalMeanToEccentricLongitudeMany
        from .orbit import _hyperbolicMeanToEccentricLongitudeMany
        # expect a Vector orbit.
        dts = np.atleast_2d(time[None, :] - orbit.t[:, None])  # (nOrbit, nTime)
        lM = orbit.lM[:, None] + orbit.meanMotion[:, None] * dts
        wBound = np.atleast_1d(orbit.a) > 0
        lE = np.empty_like(lM)
        if np.any(wBound):
            ex, ey = orbit.ex[wBound], orbit.ey[wBound]
            lE[wBound] = _ellipticalMeanToEccentricLongitudeMany(
                lM[wBound], ex, ey
            )
        if np.any(~wBound):
            ex, ey = orbit.ex[~wBound], orbit.ey[~wBound]
            lE[~wBound] = _hyperbolicMeanToEccentricLongitudeMany(
                lM[~wBound], ex, ey
            )
        r, v = orbit._rvFromEquinoctial(lE=lE)
        return r, v

    def __hash__(self):
        return hash("KeplerianPropagator")

    def __eq__(self, rhs):
        return isinstance(rhs, KeplerianPropagator)


class SeriesPropagator(Propagator):
    """ Propagate using Taylor series expansion of Keplerian motion.  This is
    quite a bit faster than the full Keplerian propagator, but only valid for
    short time intervals.  Using order=2 for a few seconds of time though is
    definitely reasonable.

    Parameters
    ----------
    order : int, optional
        Order of series expansion.  1 = constant velocity, 2 = constant
        acceleration
    """
    def __init__(self, order=2):
        self.order = order

    def __repr__(self):
        if self.order == 2:
            return "SeriesPropagator()"
        return "SeriesPropagator({})".format(self.order)

    def _getRVOne(self, orbit, time):
        dts = np.atleast_1d(time - orbit.t)
        if self.order == 0:
            # Just constant position.  Weird, but allowed.
            rs = np.broadcast_to(orbit.r, (len(time), 3))
            # Or should I use 0 here?
            vs = np.broadcast_to(orbit.v, (len(time), 3))
            return rs, vs
        if self.order == 1:
            # Constant velocity
            rs = orbit.r + orbit.v * dts[:, None]
            vs = np.broadcast_to(orbit.v, (len(time), 3))
            return rs, vs
        if self.order == 2:
            # Constant acceleration
            accel = -orbit.mu * orbit.r / norm(orbit.r)**3
            rs = orbit.r + dts[:, None] * (orbit.v + 0.5 * dts[:, None] * accel)
            vs = orbit.v + dts[:, None] * accel
            return rs, vs
        if self.order == 3:
            accel = -orbit.mu * orbit.r / norm(orbit.r)**3
            jerk = -orbit.mu * (
                orbit.r * (-3 * np.dot(orbit.r, orbit.v) / norm(orbit.r)**5) + orbit.v / norm(orbit.r)**3
            )
            rs = orbit.r + dts[:, None] * (
                orbit.v + 0.5 * dts[:, None] * (
                    accel + dts[:, None] * jerk / 3.
                )
            )
            vs = orbit.v + dts[:, None] * (accel + 0.5 * dts[:, None] * jerk)
            return rs, vs
        else:
            # I think this is basically the f and g series idea in Escobal
            # (1965) section 3.9.1
            raise NotImplementedError("order>3 not yet implemented")

    def __hash__(self):
        return hash(("SeriesPropagator", self.order))

    def __eq__(self, rhs):
        if not isinstance(rhs, SeriesPropagator):
            return False
        return self.order == rhs.order


class SGP4Propagator(Propagator):
    """Propagate using simplified perturbation model SGP4.

    Parameters
    ----------
    t : float or astropy.time.Time, optional
        Reference time at which to compute frame transformation between GCRF
        and TEME.  SGP4 calculations occur in the TEME frame, but useful input
        and output is in the GCRF frame.  In principle, one could do the
        transformation at every instant in time for which the orbit is queried.
        However, the rate of change in the transformation is small, ~0.15 arcsec
        per day, so here we just use a single transformation.

        If float, then should correspond to GPS seconds;
        i.e., seconds since 1980-01-06 00:00:00 UTC

        If None, then use the time of the orbit being propagated.
    truncate : bool, optional
        Truncate elements to precision of TLE ASCII format?  This may be
        required in order to reproduce the results of running sgp4 directly from
        a TLE.
    """
    def __init__(self, t=None, truncate=False):
        if isinstance(t, Time):
            t = t.gps
        self.t = t
        self.truncate = truncate

    def __repr__(self):
        return "SGP4Propagator()"

    def _getRVOne(self, orbit, time):
        from sgp4.api import Satrec, WGS72
        from .orbit import _ellipticalEccentricToMeanAnomaly, _ellipticalTrueToEccentricAnomaly
        from .constants import WGS72_EARTH_MU
        from .io import make_tle

        if self.truncate:
            line1, line2 = make_tle(*orbit.kozaiMeanKeplerianElements, orbit.t)
            sat = Satrec.twoline2rv(line1, line2)
        else:
            a, e, i, pa, raan, trueAnomaly = orbit.kozaiMeanKeplerianElements
            meanAnomaly = _ellipticalEccentricToMeanAnomaly(
                _ellipticalTrueToEccentricAnomaly(
                    trueAnomaly % (2 * np.pi), e
                ),
                e
            )
            meanMotion = np.sqrt(WGS72_EARTH_MU / np.abs(a**3)) * 60.0  # rad/m
            tt = Time(orbit.t, format='gps').utc
            epoch = tt.mjd - 33281.0
            sat = Satrec()
            sat.sgp4init(
                WGS72,        # gravity model
                'i',          # 'a' = old AFSPC mode, 'i' = improved mode
                0,            # satnum: Satellite number
                epoch,        # epoch: days since 1949 December 31 00:00 UT
                0.0,          # bstar: drag coefficient (kg/m2er)
                0.0,          # ndot: ballistic coefficient (revs/day)
                0.0,          # nddot: second derivative of mean motion (revs/day^3)
                e,            # ecco: eccentricity
                pa,           # argpo: argument of perigee (radians)
                i,            # inclo: inclination (radians)
                meanAnomaly,  # mo: mean anomaly (radians)
                meanMotion,   # no_kozai: mean motion (radians/minute)
                raan,         # nodeo: right ascension of ascending node (radians)
            )

        rs, vs = [], []
        for t in time:
            e, r, v = sat.sgp4_tsince((t - orbit.t) / 60.0)
            rs.append(r)
            vs.append(v)
        rs = np.array(rs)
        vs = np.array(vs)
        tref = self.t if self.t is not None else orbit.t
        rot = teme_to_gcrf(tref)
        rs = np.dot(rot, rs.T).T
        vs = np.dot(rot, vs.T).T
        rs *= 1e3  # km -> m
        vs *= 1e3  # km/s -> m/s
        return rs, vs

    def __hash__(self):
        t = self.t.gps if isinstance(self.t, Time) else self.t
        return hash(("SGP4Propagator", t))

    def __eq__(self, rhs):
        if not isinstance(rhs, SGP4Propagator):
            return False
        return np.all(self.t == rhs.t)


class SciPyPropagator(Propagator):
    """Propagate using the scipy.integrate.solve_ivp ODE solver.

    Parameters
    ----------
    accel : ssapy.Accel
        Accel object containing the acceleration model by which to propagate.
    ode_kwargs : dict
        Keyword arguments to pass to `scipy.integrate.solve_ivp`.  Of particular
        interest may be the kwarg `rtol`, which usually yields reasonable
        results when set ~1e-7.  For best results, check for convergence.
    """
    def __init__(self, accel, ode_kwargs=None):
        self.accel = accel
        if ode_kwargs is None:
            ode_kwargs = {'rtol': 1e-7}
        self.ode_kwargs = ode_kwargs

    def __repr__(self):
        return "SciPyPropagator({!r}, {!r})".format(self.accel, self.ode_kwargs)

    @staticmethod
    def _concatenateOdeSolutions(sol0, sol1):
        from scipy.integrate._ivp.common import OdeSolution
        # Ensure solutions are ascending
        if not sol0.ascending:
            sol0 = OdeSolution(sol0.ts[::-1], sol0.interpolants[::-1])
        if not sol1.ascending:
            sol1 = OdeSolution(sol1.ts[::-1], sol1.interpolants[::-1])

        # Just eliminate degenerate solutions
        if sol0.ts.size == 2 and sol0.ts[0] == sol0.ts[1]:
            return sol1
        if sol1.ts.size == 2 and sol1.ts[0] == sol1.ts[1]:
            return sol0

        # Merge non-degenerate solutions
        assert sol0.ts[-1] == sol1.ts[0]
        ts = np.concatenate([sol0.ts[:-1], sol1.ts])
        interpolants = sum([s.interpolants for s in [sol0, sol1]], [])
        return OdeSolution(ts, interpolants)

    def _solve_piecewise_ivp(self, fp, t_span, sol):
        from scipy.integrate import solve_ivp
        # solve from t_span[0] to t_span[1]
        # we need to cut this into pieces at each time in
        # self.accel.time_breakpoints.
        tbreak = self.accel.time_breakpoints
        # make infinities finite but outside the range t_span
        tbreak = np.clip(tbreak, np.min(t_span) - 1, np.max(t_span) + 1)
        tbreakind = np.arange(len(tbreak))
        t_span_ind = np.interp(t_span, tbreak, tbreakind)
        tbreak = tbreak[(tbreakind > np.min(t_span_ind)) & (tbreakind < np.max(t_span_ind))]
        if t_span[1] < t_span[0]:
            tbreak = tbreak[::-1]
        alltimes = np.concatenate([[t_span[0]], tbreak, [t_span[1]]])
        for t0, t1 in zip(alltimes[:-1], alltimes[1:]):
            soln = solve_ivp(
                fp,
                [t0, t1],
                sol(t0),
                dense_output=True,
                **self.ode_kwargs
            )
            if not soln.success:
                raise ValueError(soln.message)
            if t1 > t0:
                sol = self._concatenateOdeSolutions(sol, soln.sol)
            else:
                sol = self._concatenateOdeSolutions(soln.sol, sol)
        return sol

    def _getRVOne(self, orbit, tQuery):
        from scipy.integrate._ivp.base import ConstantDenseOutput
        from scipy.integrate._ivp.common import OdeSolution
        # Pattern for ScipyPropagator interpolant is just:
        # OdeSolution
        container = _InterpolantCache(orbit, self)

        def fp(t, s):
            r = s[0:3]
            v = s[3:6]
            return np.hstack([v, self.accel(r, v, t, **orbit.propkw)])

        tmin, tmax = np.min(tQuery), np.max(tQuery)
        update = False
        if len(container) == 0:
            ts = np.array([orbit.t, orbit.t])
            interpolants = [ConstantDenseOutput(
                orbit.t,
                orbit.t,
                np.hstack([orbit.r, orbit.v])
            )]
            sol = OdeSolution(ts, interpolants)
            update = True
        else:
            sol = container[0]
        if tmin < sol.ts[0]:
            sol = self._solve_piecewise_ivp(
                fp,
                [sol.ts[0], tmin],
                sol)
            update = True
        if tmax > sol.ts[-1]:
            sol = self._solve_piecewise_ivp(
                fp,
                [sol.ts[-1], tmax],
                sol)
            update = True
        if update:
            container.clear()
            container.append(sol)
        out = sol(tQuery).T
        return out[:, 0:3], out[:, 3:6]

    def __hash__(self):
        return hash((
            "SciPyPropagator",
            self.accel,
            frozenset(self.ode_kwargs.items())
        ))

    def __eq__(self, rhs):
        if not isinstance(rhs, SciPyPropagator):
            return False
        return (
            self.accel == rhs.accel and self.ode_kwargs == rhs.ode_kwargs
        )


class RKPropagator(Propagator, ABC):
    """
    Abstract base class for Runge-Kutta-based orbit propagators.

    The `RKPropagator` class provides a framework for implementing orbit propagation 
    using Runge-Kutta methods. Subclasses must override specific methods to define 
    the propagation logic and interpolation behavior.

    Attributes:
        _minPoints (int): Minimum number of points required for interpolation. 
            This should be overridden by subclasses.

    Methods:
        _prop():
            Abstract method that subclasses must implement to define the propagation logic.
        _getRVOne(orbit, tQuery):
            Retrieves the position and velocity of an orbit at specified query times 
            using interpolation.

    Args:
        orbit (Orbit): The orbit object containing initial conditions and propagation settings.
        tQuery (numpy.ndarray): An array of times at which the position and velocity are queried.

    Returns:
        tuple: A tuple containing:
            - r (numpy.ndarray): Position vectors at the query times.
            - v (numpy.ndarray): Velocity vectors at the query times.

    Notes:
        - The `_getRVOne` method uses an interpolation cache to store computed states and times, 
          minimizing redundant computations.
        - The Runge-Kutta propagation logic must be implemented in the `_prop` method by subclasses.
        - Spline interpolation is used for smooth querying of position and velocity data.
        - The `container` object caches the interpolant data, including times, states, and spline objects.

    Example:
        Subclass implementation:
```python
        class RK4Propagator(RKPropagator):
            _minPoints = 4

            def _prop(self, times, states, h, t_target, propkw):
                # Implement RK4 propagation logic here
                ...
```
    Raises:
        NotImplementedError: If `_prop` is not overridden by a subclass.
    """
    _minPoints = None  # subclass should override

    @abstractmethod
    def _prop(self):
        ...  # Subclasses must override

    def _getRVOne(self, orbit, tQuery):
        from collections import deque
        from scipy.interpolate import make_interp_spline
        # Pattern for RK interpolant is:
        # times, states, h_pre, h_app, spline
        container = _InterpolantCache(orbit, self)

        tmin, tmax = np.min(tQuery), np.max(tQuery)
        if len(container) == 0:
            times = deque([orbit.t])
            states = deque([np.hstack([orbit.r, orbit.v])])
            h_pre = -self.h
            h_app = self.h
            spline = None
        else:
            times, states, h_pre, h_app, spline = container
        remake_spline = False
        if times[0] >= tmin:
            h_pre = self._prop(times, states, h_pre, tmin, orbit.propkw)
            remake_spline = True
        if times[-1] <= tmax:
            h_app = self._prop(times, states, h_app, tmax, orbit.propkw)
            remake_spline = True
        if remake_spline:
            spline = make_interp_spline(times, states, k=self._minPoints)
            container.clear()
            container.extend([times, states, h_pre, h_app, spline])
        out = spline(tQuery)
        return out[:, 0:3], out[:, 3:6]


class RK4Propagator(RKPropagator):
    """Runge-Kutta 4th order numerical integrator.

    Parameters
    ----------
    accel : ssapy.Accel
        Accel object containing the acceleration model by which to propagate.
    h : float
        Step size in seconds.  Reasonable values are ~50s for LEO propagations
        over a ~day for ~meter accuracy, or 1000s for GEO propagations over a
        few days with ~meter accuracy.  For best results, check for convergence.
    """
    _minPoints = 3

    def __init__(self, accel, h):
        self.accel = accel
        self.h = h

    def __repr__(self):
        return "RK4Propagator({!r}, {!r})".format(self.accel, self.h)

    def _prop(self, times, states, h, tthresh, propkw):
        """Propagate in one direction.

        Parameters
        ----------
        times : collections.deque of float
            Known times.  Must be non-empty.  Will be mutated.
        states : collections.deque of ndarray
            Known states.  Must be non-empty.  Will be mutated.
        h : float
            time increment
        tthresh : float
            Stop when crossing this boundary
        propkw : dict
            orbit.propkw arguments

        Returns
        -------
        h : float
            final time increment
        """
        def fp(s, t):
            r = s[0:3]
            v = s[3:6]
            return np.hstack([v, self.accel(r, v, t, **propkw)])

        if h > 0:
            t = times[-1]
            state = states[-1]
            pred = lambda t: t <= tthresh
        else:
            t = times[0]
            state = states[0]
            pred = lambda t: t >= tthresh
        keepGoing = True
        while keepGoing:
            # test here so we always get 1 extra iteration...
            if not pred(t) and len(times) >= self._minPoints:
                keepGoing = False
            k1 = fp(state, t)
            k2 = fp(state + 0.5 * h * k1, t + 0.5 * h)
            k3 = fp(state + 0.5 * h * k2, t + 0.5 * h)
            k4 = fp(state + h * k3, t + h)
            state = state + h / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4)
            t = t + h
            if h > 0:
                states.append(state)
                times.append(t)
            else:
                states.appendleft(state)
                times.appendleft(t)
        return h

    def __hash__(self):
        return hash((
            "RK4Propagator",
            self.accel,
            self.h
        ))

    def __eq__(self, rhs):
        if not isinstance(rhs, RK4Propagator):
            return False
        return (
            self.accel == rhs.accel and self.h == rhs.h
        )


class RK8Propagator(RKPropagator):
    """Runge-Kutta 8th order numerical integrator.

    Parameters
    ----------
    accel : ssapy.Accel
        Accel object containing the acceleration model by which to propagate.
    h : float
        Step size in seconds.  ~70s yields accuracy of ~1e-6 meters at GEO over
        a couple of days.  ~20s yields accuracy of ~1e-5 meters at LEO over a
        few hours.
    """
    _minPoints = 7

    def __init__(self, accel, h):
        self.accel = accel
        self.h = h

    # Class level variables for Butcher tableau
    c = np.array([0, 1 / 18, 1 / 12, 1 / 8, 5 / 16, 3 / 8, 59 / 400, 93 / 200, 5490023248 / 9719169821, 13 / 20, 1201146811 / 1299019798, 1, 1])
    a = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  [1 / 18, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  [1 / 48, 1 / 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  [1 / 32, 0, 3 / 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  [5 / 16, 0, -75 / 64, 75 / 64, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                  [3 / 80, 0, 0, 3 / 16, 3 / 20, 0, 0, 0, 0, 0, 0, 0, 0],
                  [29443841 / 614563906, 0, 0, 77736538 / 692538347, -28693883 / 1125000000, 23124283 / 1800000000, 0, 0, 0, 0, 0, 0, 0],
                  [16016141 / 946692911, 0, 0, 61564180 / 158732637, 22789713 / 633445777, 545815736 / 2771057229, -180193667 / 1043307555, 0, 0, 0, 0, 0, 0],
                  [39632708 / 573591083, 0, 0, -433636366 / 683701615, -421739975 / 2616292301, 100302831 / 723423059, 790204164 / 839813087, 800635310 / 3783071287, 0, 0, 0, 0, 0],
                  [246121993 / 1340847787, 0, 0, -37695042795 / 15268766246, -309121744 / 1061227803, -12992083 / 490766935, 6005943493 / 2108947869, 393006217 / 1396673457, 123872331 / 1001029789, 0, 0, 0, 0],
                  [-1028468189 / 846180014, 0, 0, 8478235783 / 508512852, 1311729495 / 1432422823, -10304129995 / 1701304382, -48777925059 / 3047939560, 15336726248 / 1032824649, -45442868181 / 3398467696, 3065993473 / 597172653, 0, 0, 0],
                  [185892177 / 718116043, 0, 0, -3185094517 / 667107341, -477755414 / 1098053517, -703635378 / 230739211, 5731566787 / 1027545527, 5232866602 / 850066563, -4093664535 / 808688257, 3962137247 / 1805957418, 65686358 / 487910083, 0, 0],
                  [403863854 / 491063109, 0, 0, -5068492393 / 434740067, -411421997 / 543043805, 652783627 / 914296604, 11173962825 / 925320556, -13158990841 / 6184727034, 3936647629 / 1978049680, -160528059 / 685178525, 248638103 / 1413531060, 0, 0]])
    b8 = np.array([14005451 / 335480064, 0, 0, 0, 0, -59238493 / 1068277825, 181606767 / 758867731, 561292985 / 797845732, -1041891430 / 1371343529, 760417239 / 1151165299, 118820643 / 751138087, -528747749 / 2220607170, 1 / 4])

    def __repr__(self):
        return "RK8Propagator({!r}, {!r})".format(self.accel, self.h)

    def _prop(self, times, states, h, tthresh, propkw):
        """Propagate in one direction.

        Parameters
        ----------
        times : collections.deque of float
            Know times.  Must be non-empty.  Will be mutated.
        state : collections.deque of ndarray
            Known states.  Must be non-empty.  Will be mutated.
        h : float
            time increment
        tthresh : float
            Stop when crossing this boundary
        propkw : dict
            orbit.propkw arguments

        Returns
        -------
        h : float
            final time increment
        """
        # Make local to save typing, and dot lookups.
        c = self.c
        a = self.a
        b8 = self.b8

        # Can go forward or backward depending on sign of h
        def fp(s, t):
            r = s[0:3]
            v = s[3:6]
            return np.hstack([v, self.accel(r, v, t, **propkw)])

        if h > 0:
            t = times[-1]
            state = states[-1]
            pred = lambda t: t <= tthresh
        else:
            t = times[0]
            state = states[0]
            pred = lambda t: t >= tthresh
        keepGoing = True
        while keepGoing:
            # test here so we always get 1 extra iteration, which seems to
            # interpolate better
            if not pred(t) and len(times) >= self._minPoints:
                keepGoing = False
            k = np.zeros((13, 6), dtype=float)
            for i in range(13):
                k[i] = h * fp(state + np.dot(a[i], k), t + c[i] * h)
            state = state + np.dot(b8, k)
            t = t + h
            if h > 0:
                states.append(state)
                times.append(t)
            else:
                states.appendleft(state)
                times.appendleft(t)
        return h

    def __hash__(self):
        return hash((
            "RK8Propagator",
            self.accel,
            self.h
        ))

    def __eq__(self, rhs):
        if not isinstance(rhs, RK8Propagator):
            return False
        return (
            self.accel == rhs.accel and self.h == rhs.h
        )


class RK78Propagator(RK8Propagator):
    """Runge-Kutta 8th order numerical integrator with adaptive step size
    computed from embedded 7th order integrator error estimate.

    Parameters
    ----------
    accel : ssapy.Accel
        Accel object containing the acceleration model by which to propagate.
    h : float
        Initial step size in seconds.  A few 10s of seconds is usually a good
        starting point here; it'll automatically be adjusted by the algorithm.
    tol : float or array of float.
        Tolerance for a single integrator step.  Used to adaptively change the
        integrator step size.  Broadcasts to 6-dimensions.  A good target is
        usually ~[1e-6, 1e-6, 1e-6, 1e-9, 1e-9, 1e-9] for cm accuracy at GEO
        over a few days, or around LEO over a few hours.
    """
    _minPoints = 7

    def __init__(self, accel, h, tol=(1e-6,) * 3 + (1e-9,) * 3):
        self.accel = accel
        self.h = h
        self.tol = tol

    # Inherit most class vars from RK8Propagator, but need b7 coefficients
    b7 = np.array([13451932 / 455176623, 0, 0, 0, 0, -808719846 / 976000145, 1757004468 / 5645159321, 656045339 / 265891186, -3867574721 / 1518517206, 465885868 / 322736535, 53011238 / 667516719, 2 / 45, 0])

    def __repr__(self):
        return "RK78Propagator({!r}, {!r}, {!r})".format(self.accel, self.h, self.tol)

    def _prop(self, times, states, h, tthresh, propkw):
        """Propagate in one direction.

        Parameters
        ----------
        times : collections.deque of float
            Know times.  Must be non-empty.  Will be mutated.
        state : collections.deque of ndarray
            Known states.  Must be non-empty.  Will be mutated.
        h : float
            time increment
        tthresh : float
            Stop when crossing this boundary
        propkw : dict
            orbit.propkw arguments

        Returns
        -------
        h : float
            final time increment
        """
        # Make local to save typing, and dot lookups.
        c = self.c
        a = self.a
        b8 = self.b8
        b7 = self.b7

        # Can go forward or backward depending on sign of h
        def fp(s, t):
            r = s[0:3]
            v = s[3:6]
            return np.hstack([v, self.accel(r, v, t, **propkw)])

        def step(h, t, state):
            while True:
                k = np.zeros((13, 6), dtype=float)
                for i in range(13):
                    k[i] = h * fp(state + np.dot(a[i], k), t + c[i] * h)
                result7 = state + np.dot(b7, k)
                result8 = state + np.dot(b8, k)
                errmax = np.max(np.abs(result7 - result8) / self.tol)
                if errmax > (1.0):
                    h *= max(0.1, 0.9 * errmax**(-1 / 7))
                    continue
                else:
                    # reset size for next step and break
                    errmax = max(errmax, 1e-15)  # avoid underflow
                    hnext = h * min(5.0, 0.9 * errmax**(-1 / 8))
                    return h, result8, hnext

        if h > 0:
            t = times[-1]
            state = states[-1]
            pred = lambda t: t <= tthresh
        else:
            t = times[0]
            state = states[0]
            pred = lambda t: t >= tthresh
        keepGoing = True
        while keepGoing:
            # test here so we always get 1 extra iteration...
            if not pred(t) and len(times) >= self._minPoints:
                keepGoing = False
            h, state, h_next = step(h, t, state)
            t = t + h
            h = h_next
            if h > 0:
                states.append(state)
                times.append(t)
            else:
                states.appendleft(state)
                times.appendleft(t)
        return h

    def __hash__(self):
        return hash((
            "RK78Propagator",
            self.accel,
            self.h,
            self.tol
        ))

    def __eq__(self, rhs):
        if not isinstance(rhs, RK78Propagator):
            return False
        return (
            self.accel == rhs.accel and self.h == rhs.h and self.tol == rhs.tol
        )


def default_numerical(*args, cls=None, accel=None, extra_accel=None):
    """Construct a numerical propagator with sensible default acceleration.

    Parameters
    ----------
    *args : list
        Arguments to Propagator
    cls : Propagator
        class to use.  Default of None means SciPyPropagator.
    accel : Accel
        acceleration model to use.  Default of None means Earth(4, 4), sun, moon.
    extra_accel : Accel or list of Accel, optional
        Additional accelerations to add to `accel`.

    Returns
    -------
    Instance of Propagator with desired Accel model.
    """
    from .accel import AccelKepler, AccelSum
    from .gravity import AccelHarmonic, AccelThirdBody
    from .body import get_body
    if accel is None:
        earth = get_body("earth")
        aK = AccelKepler(earth.mu)
        aH = AccelHarmonic(earth, 4, 4)
        aSun = AccelThirdBody(get_body("sun"))
        aMoon = AccelThirdBody(get_body("moon"))
        accellist = [aK, aH, aSun, aMoon]
        if extra_accel is not None:
            if not isinstance(extra_accel, list):
                extra_accel = [extra_accel]
            accellist += extra_accel
        accel = AccelSum(accellist)
    if cls is None:
        from functools import partial
        cls = partial(SciPyPropagator, ode_kwargs=dict(method='DOP853', rtol=1e-9, atol=(1e-1, 1e-1, 1e-1, 1e-4, 1e-4, 1e-4)))
    return cls(accel, *args)



import json

def update_documentation():
    # Load benchmark results
    with open("benchmark_results_frame.json", "r") as f:
        results = json.load(f)

    # Format the results for documentation
    formatted_results = "\nBenchmark Results:\n------------------\n"
    for result in results:
        formatted_results += f"Test Name: {result['test_name']}\n"
        formatted_results += f"Execution Time: {result['execution_time']} seconds\n"
        formatted_results += f"Status: {result['status']}\n\n"

    # Append results to benchmarks.rst
    doc_path = "docs/source/benchmarks.rst"
    with open(doc_path, "a") as doc_file:
        doc_file.write(formatted_results)

if __name__ == "__main__":
    update_documentation()


"""
Classes for sampling orbit model parameters.
"""

import numpy as np
from .orbit import Orbit as _Orbit
from .compute import rv
from . import utils


class Particles:
    """A class for importance (re-)sampling orbit model parameter samples

    Parameters
    ----------
    particles : (num_particles, 6) array_like
        Positions and velocities of orbiting object at epoch in meters and meters / second.
        The 'chain' part of the output from rvsampler.XXSampler.sample
    rvprobability : class instance
        An instance of an RVProbability for this epoch. The Particles class wraps the `lnlike`
        method of the supplied RVProbability object.
    ln_weights : (num_particles,) array_like, optional
        Weights for each particle in the input particles

    Attributes
    ----------
    particles : (num_particles, ) array_like
        Positions and velocities of orbiting object at epoch in meters and meters / second.
        The 'chain' part of the output from RVSampler.sample
    rvprobability : class instance
        An instance of an RVProbability for this epoch. The Particles class wraps the `lnlike`
        method of the supplied RVProbability object.
    ln_wts : (num_particles,) array_like
        Log weights for each particle.
    num_particles : int
        Number of particles
    initial_particles : (num_particles, 6) array_like
        Copy of input particles to enable reset_to_pseudo_prior()
    initial_ln_wts : (num_particles,) array_like
        Copy of input log weights to enable reset_to_pseudo_prior()
    epoch : float or astropy.time.Time
        The time at which the model parameters (position and velocity) are specified. If float,
        then should correspond to GPS seconds; i.e., seconds since 1980-01-06 00:00:00 UTC.
    orbits : list
        List of Orbit class instances derived from each particle
    lnpriors : (num_particles,) array_like
        Log prior probabilities for each particle. The priors are as set in the input RVProbability
        object.

    Methods
    -------
    reset_to_pseudo_prior()
        Reset the particles and weights to their values at instantiation
    move(epoch)
        Move particles to the specified epoch
    lnlike(epoch_particles)
        Evaluate the log-likelihood of the input particles given internally stored measurements
    reweight(epoch_particles)
        Reweight particles using cross-epoch likelihoods
    resample(num_particles)
        Resample particles to achieve near-uniform weighting per particle
    fuse(epoch_particles, verbose)
        Fuse particles from a different epoch and fuse internal particles and weights
    draw_orbit()
        Draw a single orbit model from the internal list of particles
    mean()
        Evaluate the weighted mean of the particle values
    """
    def __init__(self,
                 particles,
                 rvprobability,
                 lnpriors=None,
                 ln_weights=None):
        self._orbits = None
        self._lnpriors = None
        # Check if we have 3D array of particles from RVSampler.
        # If so, flatten into 2D array of (samples, parameters)
        if particles.ndim == 3:
            particles = np.vstack(particles)
            if lnpriors is not None:
                if lnpriors.ndim != 2:
                    raise ValueError("Supplied lnpriors do not match expected shape from particles")
                self._lnpriors = lnpriors.ravel()
        self.particles = particles
        self.rvprobability = rvprobability
        if ln_weights is None:
            # ln_weights = np.zeros(particles.shape[0], dtype=np.float64)
            # Subtract the log-prior probabilities to approximate samples from the likelihood
            ln_weights = -self.lnpriors
        self.ln_wts = ln_weights
        self.num_particles = self.particles.shape[0]
        #
        # Save a copy of the particles and weights to enable resets to
        # pseudo-prior
        #
        self.initial_particles = particles.copy()
        self.initial_ln_wts = ln_weights.copy()

    def __repr__(self):
        pmean = self.mean()
        out = "Particles(r={!r}, v={!r}, t={!r}".format(pmean[0:3], pmean[3:6], self.epoch)
        out += ")"
        return out

    def reset_to_pseudo_prior(self):
        self.particles = self.initial_particles.copy()
        self.ln_wts = self.initial_ln_wts.copy()

    @property
    def epoch(self):
        """ Get the epoch time at which the particles (i.e., position and velocity) are defined
        """
        return self.rvprobability.epoch

    @property
    def orbits(self):
        """ Get a list of orbits corresponding to each stored particle
        """
        if self._orbits is None:
            self._orbits = [_Orbit(p[0:3], p[3:6], t=self.epoch) for p in self.particles]
        return self._orbits

    @property
    def lnpriors(self):
        """ Get an array of log prior probabilities for each particle
        """
        if self._lnpriors is None:
            self._lnpriors = np.array([self.rvprobability.lnprior(orbit) for orbit in self.orbits])
        return self._lnpriors

    def move(self, epoch):
        """ Move particles to the given epoch

        Parameters
        ----------
        epoch : astropy.time.Time

        Returns
        -------
        particles : array (num_particles, 6)
            Position and velocity for each particle at the new epoch
        """
        r, v = rv(self.orbits, epoch)
        return np.column_stack((r, v))

    def lnlike(self, orbits):
        """ Evaluate this epoch likelihood given one or more particles

        Parameters
        ----------
        orbits : List of Orbit class instances
            A list of the Orbit models for each particle, e.g., as generated by Particles.orbits

        Returns
        -------
        lnlike : array (num_particles, )
            Array of log-likelihood values for each particle
        """
        return np.array([self.rvprobability.lnlike(orbit) for orbit in orbits])

    def reweight(self, epoch_particles):
        """ Reweight particles using cross-epoch likelihoods

        The weights for the kth particle are,
            w_k = log(Prod_i L_(d_i | theta_k) / L(d_j | theta_k))
                = Sum_{i /= j} log(L(d_i | theta_k))
        where j is the current epoch, theta are the particle parameters, and d_i is the data from
        epoch i.

        Parameters
        ----------
        epoch_particles : Particles
            The particles object from another observation

        Returns
        -------
        status : bool
            Status (True means success). Side effect is to update the internal particles state.
        """
        status = True

        # Update particle weights for the input epoch given this epoch likelihood
        epoch_ln_wts = self.lnlike(epoch_particles.orbits) + epoch_particles.lnpriors
        epoch_ln_wts += epoch_particles.ln_wts

        # Update particle weights for this epoch given input epoch likelihood
        ln_wts = epoch_particles.lnlike(self.orbits) + self.lnpriors
        ln_wts += self.ln_wts

        # Append particles and weights in anticipation of resampling / downsampling
        self.particles = np.row_stack((self.particles, epoch_particles.move(self.epoch)))
        self.ln_wts = np.append(ln_wts, epoch_ln_wts)

        if np.logaddexp.reduce(self.ln_wts) < -100.:
            status = False
        return status

    def resample(self, num_particles):
        """ Resample particles to achieve near-uniform weighting per particle

        Parameters
        ----------
        num_particles : int
            The number of particles to keep

        Returns
        -------
            None. Side effect is to update the internal particles state.
        """
        # TODO: Decide if resampling is really required here?
        self.particles, wts = utils.resample(self.particles,
                                             self.ln_wts, pod=True)
        self.ln_wts = np.log(wts)

        # Reset the stored log priors array
        self._lnpriors = None
        # Reset the stored orbits array
        self._orbits = None

        # Down-sample (without replacement) from the current particle list if
        # the requested number of particles does not match the current number.
        if self.particles.shape[0] > self.num_particles:
            ndx = np.random.choice(self.particles.shape[0], size=num_particles, replace=False)
            self.particles = self.particles[ndx, ]
            self.ln_wts = self.ln_wts[ndx]
        elif self.num_particles > self.particles.shape[0]:
            raise ValueError("Requested more particles than we have")
        return None

    def fuse(self, epoch_particles, verbose=False):
        """ Add particles from a new epoch and fuse all epoch particles and weights accordingly

        Parameters
        ----------
        epoch_particles : Particles
            The particles object from another observation
        verbose : bool, optional
            Enable verbose output to stdout

        Returns
        -------
            None. Side effect is to fuse the internal particles state.
        """
        num_part_out = self.num_particles
        # Update the weight values by 'cross-pollinating' epoch likelihoods
        status = self.reweight(epoch_particles)
        # Check that we have at least some non-zero weights for particles
        if status > 0 and verbose:
            print("All weights are negligible in Particles class")
            print("\tlog-norm:",np.logaddexp.reduce(self.ln_wts))
        # Resample the weights to maintain only significant particles
        self.resample(num_particles=num_part_out)
        return None

    def draw_orbit(self):
        """ Draw a single orbit model from the internal particle collection with probability
        proportional to the internal weights.

        Returns
        -------
        particle : array (1, 6)
            Position and velocity for one particle
        """
        prob = utils.get_normed_weights(self.ln_wts)
        particle_ndx = np.random.choice(self.num_particles, size=1, p=prob)
        p = self.particles[particle_ndx, :][0]
        return [_Orbit(p[0:3], p[3:6], t=self.epoch)]

    def mean(self):
        """ Evaluate the weighted mean of all particles

        Returns
        -------
        mean : (6,) array_like
            Weighted mean of the particle values (3*m, 3*m/s)
        """
        return np.average(self.particles, axis=0, weights=np.exp(self.ln_wts))



"""
Tools for orbital dynamics, satellite tracking, and celestial mechanics.
"""

import numpy as np
from astropy.time import Time
import astropy.units as u

from .body import get_body
from .constants import RGEO, LD, EARTH_RADIUS, EARTH_MU, MOON_RADIUS, MOON_MU
from .propagator import KeplerianPropagator
from .utils import (
    norm, normed, unitAngle3, get_angle, LRU_Cache, lb_to_unit, sunPos, _gpsToTT,
    iers_interp, rotation_matrix_from_vectors, angle_between_vectors, gcrf_to_itrf
)
from .orbit import Orbit
from .ellipsoid import Ellipsoid

try:
    import erfa
except ImportError:
    # Let this raise
    import astropy._erfa as erfa


def _doSqueeze(squeezeOrbit, squeezeTime, *args):
    out = []
    for i, arg in enumerate(args):
        if squeezeTime:
            if squeezeOrbit:
                out.append(arg[0, 0])  # scalar
            else:
                out.append(np.squeeze(arg, axis=1))  # (nOrbit,)
        else:
            if squeezeOrbit:
                out.append(np.squeeze(arg, axis=0))  # (nTime,)
            else:
                out.append(arg)  # (nOrbit, nTime)
    if len(out) == 1:
        return out[0]
    return tuple(out)


def _countOrbit(orbit):
    # orbit is one of:
    # 1) scalar Orbit
    # 2) vector Orbit
    # 3) list of scalar Orbit
    # convert to (2), set nOrbit, squeezeOrbit, and orbit.

    # check 1) and 2)
    if isinstance(orbit, Orbit):
        # check shape of r to decide scalar vs vector
        if orbit.r.ndim == 1:  # scalar Orbit
            nOrbit = 1
            squeezeOrbit = True
            if 'kozaiMeanKeplerianElements' in orbit.__dict__:
                kMKE = orbit.__dict__['kozaiMeanKeplerianElements']
            else:
                kMKE = None
            orbit = Orbit(
                np.atleast_2d(orbit.r),
                np.atleast_2d(orbit.v),
                np.atleast_1d(orbit.t),
                mu=orbit.mu,
                propkw={k: np.atleast_1d(v) for k, v in orbit.propkw.items()},
            )
            # Copying just kozaiMeanKeplerianElements for now, though maybe
            # ought to retain other (all?) lazy_properties attributes too?
            if kMKE is not None:
                orbit.kozaiMeanKeplerianElements = kMKE
        else:  # vector Orbit
            nOrbit = orbit.r.shape[0]
            squeezeOrbit = False
    else:
        import warnings
        warnings.warn(
            "list of Orbit syntax is deprecated.  "
            "Please use a vector Orbit instead.",
            DeprecationWarning
        )
        # list of scalar orbit
        nOrbit = len(orbit)
        squeezeOrbit = False
        # assumes all orbits have the same keywords and mu!
        propkeys = [] if nOrbit == 0 else orbit[0].propkw.keys()
        mu = None if nOrbit == 0 else orbit[0].mu
        orbit = Orbit(
            np.array([o.r for o in orbit]),
            np.array([o.v for o in orbit]),
            np.array([o.t for o in orbit]),
            mu=mu,
            propkw={k: np.array([o.propkw[k] for o in orbit])
                    for k in propkeys}
        )
    return nOrbit, squeezeOrbit, orbit


def _countTime(time):
    if isinstance(time, Time):
        time = time.gps
    squeezeTime = False
    try:
        nTime = len(time)
    except TypeError:
        time = np.array([time])
        nTime = 1
        squeezeTime = True
    return nTime, squeezeTime, time


def _countR(r):
    # orbit is one of:
    # 1) scalar r
    # 2) vector r
    # 3) list of scalar Orbit
    # convert to (2), set nOrbit, squeezeOrbit, and orbit.
    squeezeR = False
    if np.shape(r)[-1] == 3:
        pass
    else:
        raise ValueError(f"Incorrect r dimensions. Expected shape (n, 3), but got {np.shape(r)}.")
    # check 1) and 2)
    if r.ndim < 3:  # scalar r
        nR = 1
        r = np.reshape(np.atleast_3d(r), (nR, np.shape(r)[0], np.shape(r)[1]))
        squeezeR = True
    else:
        nR = np.shape(r)[0]
    return nR, squeezeR, r


def _processObserver(
    observer, obsPos, obsVel, nTime, squeezeTime, time, doObsVel=False
):
    # We always need obsPos output, but only sometimes obsVel.  So only require
    # obsVel if doObsVel is True.  (but always return it, and always compute it
    # if what we got as input was in an observer).
    squeezeObsPos = False
    if observer is None:
        if obsPos is None:
            raise ValueError(
                "Exactly one of obsPos and observer must be specified"
            )
        if doObsVel and obsVel is None:
            raise ValueError(
                "Exactly one of obsVel and observer required for doObsVel"
            )

    if observer is not None:
        if obsPos is not None:
            raise ValueError(
                "Exactly one of obsPos and observer must be specified"
            )
        if doObsVel and obsVel is not None:
            raise ValueError(
                "Exactly one of obsVel and observer required for doObsVel"
            )
        try:
            nObservers = len(observer)
        except TypeError:
            observer = [observer]
            nObservers = 1
            squeezeObsPos = squeezeTime
        # Now generate obsPos
        if nObservers == nTime:
            obsPos = np.empty((nObservers, 3), dtype=float)
            obsVel = np.empty((nObservers, 3), dtype=float)
            for i, (obs, t) in enumerate(zip(observer, time)):
                obsPos[i], obsVel[i] = obs.getRV(t)
        else:
            if nObservers == 1:
                obsPos, obsVel = observer[0].getRV(time)
            elif nTime == 1:
                obsPos = np.empty((nObservers, 3), dtype=float)
                obsVel = np.empty((nObservers, 3), dtype=float)
                for i, obs in enumerate(observer):
                    obsPos[i], obsVel[i] = obs.getRV(time[0])
            else:
                raise ValueError("observer and time must be broadcastable")

    if doObsVel:
        obsPos, obsVel = np.broadcast_arrays(obsPos, obsVel)
    obsPos = np.atleast_1d(obsPos)
    if obsPos.ndim == 2:
        nObsPos = obsPos.shape[0]
    else:
        obsPos = np.atleast_2d(obsPos)
        nObsPos = 1
        squeezeObsPos = True
    # times and obsPos must align
    if nObsPos != nTime:
        if nObsPos == 1:
            obsPos = np.broadcast_to(obsPos, (nTime, 3))
            if doObsVel:
                obsVel = np.broadcast_to(obsVel, (nTime, 3))
            nObsPos = nTime
        elif nTime == 1:
            time = np.broadcast_to(time, (nObsPos,))
            nTime = nObsPos
        else:
            raise ValueError("obsPos and time must be broadcastable")
    squeezeTime &= squeezeObsPos
    return obsPos, obsVel, time, squeezeTime


class HashableArrayContainer:
    """
    A container for NumPy arrays that makes them hashable and immutable.

    This class wraps a NumPy array and ensures that it is immutable by setting
    its `writeable` flag to `False`. It also provides implementations for
    `__hash__` and `__eq__`, enabling the wrapped array to be used as a key in
    hash-based data structures like dictionaries and sets.

    Attributes:
        arr (numpy.ndarray): The wrapped NumPy array, which is immutable.

    Methods:
        __hash__():
            Returns a hash value for the array based on its byte representation.
        
        __eq__(rhs):
            Compares the wrapped array with another `HashableArrayContainer` instance
            for equality based on element-wise comparison.
    
    Examples:
        >>> import numpy as np
        >>> arr1 = np.array([1, 2, 3])
        >>> arr2 = np.array([1, 2, 3])
        >>> container1 = HashableArrayContainer(arr1)
        >>> container2 = HashableArrayContainer(arr2)
        >>> container1 == container2
        True
        >>> hash(container1) == hash(container2)
        True
        >>> my_dict = {container1: "value"}
        >>> my_dict[container2]
        'value'
    """

    def __init__(self, arr):
        self.arr = arr
        self.arr.flags.writeable = False

    def __hash__(self):
        return hash(self.arr.data.tobytes())

    def __eq__(self, rhs):
        return np.all(self.arr == rhs.arr)


def rv(orbit, time, propagator=KeplerianPropagator()):
    """Calculate positions and velocities on the outer product of all supplied
    orbits and times.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Desired orbit(s)
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    propagator : Propagator, optional
        The propagator instance to use.

    Notes
    -----
    If orbit or time is scalar valued (as opposed to a list of Orbit, e.g.) then
    the corresponding dimension of the output will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    Returns
    -------
    r : array_like (n, m, 3)
        Position in meters.
    v : array_like (n, m, 3)
        Velocity in meters per second.
    """
    nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)
    nTime, squeezeTime, time = _countTime(time)

    outR, outV = _rv(orbit, HashableArrayContainer(time), propagator)

    return _doSqueeze(squeezeOrbit, squeezeTime, outR, outV)


def __rv(orbit, time, propagator):
    outR, outV = propagator._getRVMany(orbit, time.arr)

    return outR, outV


_rv = LRU_Cache(__rv, maxsize=16)


def groundTrack(orbit, time, propagator=KeplerianPropagator(), format='geodetic'):
    """Calculate satellite ground track on the outer product of all supplied times and
    state vectors or orbits.

    Parameters
    ----------
    r : array_like (n,3) Position (m)

    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    or

    orbit : r : array_like (n,3) Position (m) or
        Orbit or list of Orbit (n,)
        Desired orbit(s)
    propagator : Propagator, optional
        The propagator instance to use.
    format : 'geodetic' or 'cartesian'
        If 'geodetic', then returns longitude, latitude, height.
        If 'cartesian', then returns xyz in ITRF frame.

    Notes
    -----
    If orbit or time is scalar valued (as opposed to a list of Orbit, e.g.) then
    the corresponding dimension of the output will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    Returns
    -------
    lon, lat, height : array_like (n, m, 3)
        Radians and meters.

    or

    x, y, z : array_like(n, m, 3)
        Meters.
    """
    if format not in ['cartesian', 'geodetic']:
        raise ValueError("Format must be either 'cartesian' or 'geodetic'")

    nTime, squeezeTime, time = _countTime(time)
    if isinstance(orbit, Orbit):
        nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)
        r, v = rv(orbit, time, propagator=propagator)  # (n, m, 3)
    else:
        nOrbit, squeezeOrbit, r = _countR(orbit)  # (n, m, 3)

    # Reverse the math in EarthObserver.getRV
    mjd_tt = _gpsToTT(time)
    d_ut1_tt_mjd, pmx, pmy = iers_interp(time)
    pn = erfa.pnm80(2400000.5, mjd_tt)
    gst = erfa.gst94(2400000.5, mjd_tt + d_ut1_tt_mjd)
    cg, sg = np.cos(gst), np.sin(gst)
    gstMat = np.zeros([len(time), 3, 3], dtype=float)
    gstMat[:, 0, 0] = cg
    gstMat[:, 0, 1] = sg
    gstMat[:, 1, 0] = -sg
    gstMat[:, 1, 1] = cg
    gstMat[:, 2, 2] = 1.0

    polar = np.zeros([len(time), 3, 3], dtype=float)
    polar[:] = np.eye(3)

    polar[:, 0, 2] = pmx
    polar[:, 1, 2] = -pmy
    polar[:, 2, 0] = -pmx
    polar[:, 2, 1] = pmy

    U = gstMat @ pn
    itrs = np.einsum(
        "tab,ntb->nta",
        (polar @ U),
        r
    )
    x = itrs[..., 0]
    y = itrs[..., 1]
    z = itrs[..., 2]
    if format == 'cartesian':
        return _doSqueeze(squeezeOrbit, squeezeTime, x, y, z)
    elif format == 'geodetic':
        ellipsoid = Ellipsoid()
        lon, lat, height = ellipsoid.cartToSphere(x, y, z)
        return _doSqueeze(squeezeOrbit, squeezeTime, lon, lat, height)


def _obsAngleCorrection(
    r, v, obsPos, obsVel, orbit, time, propagator, correctionType, max_iter=10
):
    if correctionType == "linear":
        # First order correction is to back up satellite by v * dt
        # where dt is the approximate time delay from the 0th order distance.
        # Important: *don't* do anything to the observer position, since we
        # really did do the observation at the originally given time and place.
        dt = norm(r - obsPos) / 299792458
        r -= np.einsum("nmi,nm->nmi", v, dt)
        # no change to v since linear assumption _is_ that v is constant for the
        # light time duration.
    elif correctionType == "exact":
        # Solve |r(t-dt) - r_obs(t)| = c dt for dt.
        dist = norm(r - obsPos)
        dt, dt_previous = dist / 299792458, np.inf
        iter = 0
        while np.any(np.abs(dt - dt_previous)) > 1e-12:  # picosecond accurate
            if iter > max_iter:
                raise RuntimeError(
                    "Exact light time correction did not converge in "
                    "{} iterations".format(iter)
                )
            # Unfortunately, can't just do
            # r, v = rv(orbit, time-dt, propagator=propagator)
            # because ssapy.rv isn't parallelized for n orbits and n, m times.
            # so do one orbit at a time.
            r = np.empty((len(orbit), len(time), 3))
            v = np.empty((len(orbit), len(time), 3))
            for i, o in enumerate(orbit):
                r[i], v[i] = rv(o, time - dt[i], propagator=propagator)
            dist = norm(r - obsPos)
            dt, dt_previous = dist / 299792458, dt
            iter += 1
    else:
        raise ValueError(
            "Invalid value for correctionType: {}"
            .format(correctionType)
        )
    # Apply velocity correction (i.e., diurnal aberration for an Earth-based
    # observer, and whatever you call the analogous correction for an orbital
    # observer).
    dr = norm(r - obsPos)
    r += obsVel * dr[..., None] / 299792458
    return r, v


def dircos(
    orbit, time, obsPos=None, obsVel=None, observer=None,
    propagator=KeplerianPropagator(), obsAngleCorrection=None
):
    """ Calculate observed direction-cosines of orbiting objects as viewed at
    specified times and positions.

    The direction cosines are the cosines of the angles between the vector
    pointing towards the orbiting object and the x, y, z axes.  An equivalent
    description is that they are the components of the unit vector pointing
    towards the orbiting object.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Orbit(s) for which to calculate direction cosines.
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    obsPos : array_like (m, 3), optional
        Position of observer at given time.
    obsVel : array_like (m, 3), optional
        Velocity of observer at given time.  Only required if correcting for
        diurnal aberration.
    observer : Observer or list of Observers (m,), optional
        Observer(s) used to calculate obsPos and obsVel.
    propagator : Propagator, optional
        The propagator instance to use.
    obsAngleCorrection : {None, "linear", "exact"}, optional
        Correct actual angle to observed angle, meaning account for light-time
        delay, and aberration due to the observer's velocity.  None means don't
        do any correction.  "linear" means do an aberration correction and first
        order light-time correction.  "exact" means do an aberration correction
        and iteratively solve for the exact light-time correction.  (The
        "linear" correction is almost always sufficiently accurate).

    Notes
    -----
    Exactly 1 of `obsPos` and `observer` must be supplied.  `observer` and
    `obsPos` follow similar broadcasting rules as detailed below explicitly only
    for `obsPos`.

    The length of `time` and `obsPos` must match or be broadcastable to match.
    If `orbit` is scalar-valued (an Orbit instead of a list of Orbit), then that
    dimension will be squeezed out in the return value.  Likewise, if both
    `time` and `obsPos` are scalar, that dimension will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    When doing light time corrections, the time argument is the arrival time of
    the photons at the observer, as opposed to the emission time at the
    satellite.

    Returns
    -------
    dircos : array_like (n, m, 3)
        Direction cosines on the outer product of orbit(s) and time/obsPos.
    """
    nTime, squeezeTime, time = _countTime(time)
    nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)
    obsPos, obsVel, time, squeezeTime = _processObserver(
        observer, obsPos, obsVel, nTime, squeezeTime, time,
        doObsVel=obsAngleCorrection
    )

    # At this point, should have len(orbit) == n
    # and len(obsPos) == len(time) == nTime
    # So get positions of orbits...
    r, v = rv(orbit, time, propagator=propagator)  # (n, m, 3)
    if obsAngleCorrection:
        r, v = _obsAngleCorrection(
            r, v, obsPos, obsVel, orbit, time, propagator, obsAngleCorrection
        )
    dc = normed(r - obsPos)

    return _doSqueeze(squeezeOrbit, squeezeTime, dc)


def radec(
    orbit, time, obsPos=None, obsVel=None, observer=None,
    propagator=KeplerianPropagator(), obsAngleCorrection=None, rate=False
):
    """ Calculate observed right ascension, declination, and slant range of
    orbiting objects as viewed at specified times and positions.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Orbit(s) for which to calculate ra and dec.
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    obsPos : array_like (m, 3), optional
        Positions of observers at given times.
    obsVel : array_like (m, 3), optional
        Velocity of observers at given times.
    observer : Observer or list of Observers (m,), optional
        Observer(s) used to calculate obsPos.
    propagator : Propagator, optional
        The propagator instance to use.
    rate : bool
        If True, return additionally the time derivatives of the quantity, times
        cos(dec) in the case of right ascension.
    obsAngleCorrection : {None, "linear", "exact"}, optional
        Correct actual angle to observed angle, meaning account for light-time
        delay, and aberration due to the observer's velocity.  None means don't
        do any correction.  "linear" means do an aberration correction and first
        order light-time correction.  "exact" means do an aberration correction
        and iteratively solve for the exact light-time correction.  (The
        "linear" correction is almost always sufficiently accurate).

    Notes
    -----
    Exactly 1 of `obsPos` and `observer` must be supplied.  `observer` and
    `obsPos` follow similar broadcasting rules as detailed below explicitly only
    for `obsPos`.

    The length of `time` and `obsPos` must match or be broadcastable to match.
    If `orbit` is scalar-valued (an Orbit instead of a list of Orbit), then that
    dimension will be squeezed out in the return value.  Likewise, if both
    `time` and `obsPos` are scalar, that dimension will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    When doing light time corrections, the time argument is the arrival time of
    the photons at the observer, as opposed to the emission time at the
    satellite.

    Returns
    -------
    ra, dec : array_like (n, m)
        Right ascension and declination in radians.
    range : array_like (n, m)
        (Slant) range in meters.
    If rate, also:
    raRate : array_like (n, m)
        Time derivatives of right ascension times cos(dec), rad / sec.
    decRate : array_like (n, m)
        Time derivative of declination, rad / sec.
    rangeRate : array_like (n, m)
        Time derivative of range, meter / s.
    """
    nTime, squeezeTime, time = _countTime(time)
    nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)
    obsPos, obsVel, time, squeezeTime = _processObserver(
        observer, obsPos, obsVel, nTime, squeezeTime, time,
        doObsVel=obsAngleCorrection
    )

    if rate and obsVel is None:
        raise ValueError('obsVel must not be None if rate is True.')

    # At this point, should have len(orbit) == n
    # and len(obsPos) == len(time) == nTime
    # So get positions of orbits...
    r, v = rv(orbit, time, propagator=propagator)  # (n, m, 3)
    if obsAngleCorrection:
        r, v = _obsAngleCorrection(
            r, v, obsPos, obsVel, orbit, time, propagator, obsAngleCorrection
        )

    ra, dec, slantRange, raRate, decRate, rangeRate = rvObsToRaDecRate(
        r, v, obsPos, obsVel
    )
    res = (ra, dec, slantRange)
    if rate:
        res = res + (raRate, decRate, rangeRate)

    return _doSqueeze(*((squeezeOrbit, squeezeTime) + res))


def altaz(
    orbit, time, observer, propagator=KeplerianPropagator(),
    obsAngleCorrection=None
):
    """ Calculate observed altitude and azimuth of orbiting objects as viewed at
    specified times and locations.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Orbits for which to calculate direction cosines.
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    observer : EarthObserver
        Where on Earth to compute alt and az.
    propagator : Propagator, optional
        The propagator instance to use.
    obsAngleCorrection : {None, "linear", "exact"}, optional
        Correct actual angle to observed angle, meaning account for light-time
        delay, and aberration due to the observer's velocity.  None means don't
        do any correction.  "linear" means do an aberration correction and first
        order light-time correction.  "exact" means do an aberration correction
        and iteratively solve for the exact light-time correction.  (The
        "linear" correction is almost always sufficiently accurate).

    Notes
    -----
    If `orbit` is scalar-valued (an Orbit instead of a list of Orbit), then that
    dimension will be squeezed out in the return value.  Likewise, if `time` is
    scalar, that dimension will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    When doing light time corrections, the time argument is the arrival time of
    the photons at the observer, as opposed to the emission time at the
    satellite.

    Returns
    -------
    alt, az : array_like (n, m)
        Altitude and azimuth in radians.
    """
    from astropy.coordinates import SkyCoord, AltAz, GCRS

    nTime, squeezeTime, time = _countTime(time)
    nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)
    obsPos, obsVel = observer.getRV(time)

    r, v = rv(orbit, time, propagator=propagator)  # (n, m, 3)
    if obsAngleCorrection:
        r, v = _obsAngleCorrection(
            r, v, obsPos, obsVel, orbit, time, propagator, obsAngleCorrection
        )

    # Using astropy backend for AltAz computation
    sc = SkyCoord(
        x=r[..., 0], y=r[..., 1], z=r[..., 2],
        unit='m',
        representation_type='cartesian',
        frame=GCRS(obstime=Time(time, format='gps'))
    )
    aa = sc.transform_to(AltAz(location=observer._location))

    return _doSqueeze(squeezeOrbit, squeezeTime, aa.alt.radian, aa.az.radian)


def quickAltAz(
    orbit, time, observer, propagator=KeplerianPropagator(),
    obsAngleCorrection=None
):
    """ Quickly estimate observed altitude and azimuth of orbiting objects as
    viewed at specified times and locations.

    This algorithm approximates "up" as pointing directly away from the center
    of the Earth, instead of normal to the reference ellipsoid.  Use `altAz` if
    you want values wrt the reference ellipsoid.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Orbits for which to calculate direction cosines.
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    observer : EarthObserver
        Where on Earth to compute alt and az.
    propagator : Propagator, optional
        The propagator instance to use.
    obsAngleCorrection : {None, "linear", "exact"}, optional
        Correct actual angle to observed angle, meaning account for light-time
        delay, and aberration due to the observer's velocity.  None means don't
        do any correction.  "linear" means do an aberration correction and first
        order light-time correction.  "exact" means do an aberration correction
        and iteratively solve for the exact light-time correction.  (The
        "linear" correction is almost always sufficiently accurate).

    Notes
    -----
    If `orbit` is scalar-valued (an Orbit instead of a list of Orbit), then that
    dimension will be squeezed out in the return value.  Likewise, if `time` is
    scalar, that dimension will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    When doing light time corrections, the time argument is the arrival time of
    the photons at the observer, as opposed to the emission time at the
    satellite.

    Returns
    -------
    alt, az : array_like (n, m)
        Altitude and azimuth in radians.
    """
    nTime, squeezeTime, time = _countTime(time)
    nOrbit, squeezeOrbit, orbit = _countOrbit(orbit)

    ro, vo = observer.getRV(time)  # (m, 3)
    r, v = rv(orbit, time, propagator=propagator)  # (n, m, 3)
    if obsAngleCorrection:
        r, v = _obsAngleCorrection(
            r, v, ro, vo, orbit, time, propagator, obsAngleCorrection
        )

    up = normed(ro)
    east = normed(vo)
    north = np.cross(up, east)

    dr = r - ro
    northing = np.einsum("nmi,mi->nm", dr, north)
    easting = np.einsum("nmi,mi->nm", dr, east)

    alt = np.pi / 2 - unitAngle3(
        normed(np.broadcast_to(ro, r.shape)), normed(dr)
    )
    az = np.arctan2(easting, northing)
    az %= 2 * np.pi

    return _doSqueeze(squeezeOrbit, squeezeTime, alt, az)


def radecRate(
    orbit, time, obsPos=None, obsVel=None, observer=None,
    propagator=KeplerianPropagator(),
    obsAngleCorrection=None
):
    """Calculate ra/dec rate and slant range rate of orbit at specified times
    and observer positions and velocities.

    DEPRECATED.  Use radec(..., rate=True) in new code.

    Parameters
    ----------
    orbit : Orbit or list of Orbit (n,)
        Orbit(s) for which to calculate slant range rate.
    time : array_like or astropy.time.Time (m,)
        If float (array), then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    obsPos : array_like (m, 3), optional
        Positions of observers at given times.
    obsVel : array_like (m, 3), optional
        Velocity of observers at given times.
    observer : Observer or list of Observers (m,), optional
        Observer(s) used to calculate obsPos and obsVel.
    propagator : Propagator, optional
        The propagator instance to use.
    obsAngleCorrection : {None, "linear", "exact"}, optional
        Correct actual angle to observed angle, meaning account for light-time
        delay, and aberration due to the observer's velocity.  None means don't
        do any correction.  "linear" means do an aberration correction and first
        order light-time correction.  "exact" means do an aberration correction
        and iteratively solve for the exact light-time correction.  (The
        "linear" correction is almost always sufficiently accurate).

    Notes
    -----
    Exactly 1 of `obsPos` and `observer` must be supplied.  `observer` and
    `obsPos` follow similar broadcasting rules as detailed below explicitly only
    for `obsPos`.  If `obsPos` is specified, `obsVel` must also be specified and
    congruent to `obsPos`.

    The length of `time` and `obsPos` must match or be broadcastable to match.
    If `orbit` is scalar-valued (an Orbit instead of a list of Orbit), then that
    dimension will be squeezed out in the return value.  Likewise, if both
    `time` and `obsPos` are scalar, that dimension will be squeezed out.

    For Keplerian orbit propagation it is more efficient to use a "vector Orbit"
    instead of a list of single scalar Orbits.

    When doing light time corrections, the time argument is the arrival time of
    the photons at the observer, as opposed to the emission time at the
    satellite.

    Returns
    -------
    ra : array_like (n, m)
        right ascension in radians
    raRate : array_like (n, m)
        Rate of change of right ascension*cos(dec) in radians per second.
    dec : array_link (n, m)
        declination in radians
    decRate : array_like (n, m)
        Rate of change of declination in radians per second.
    slantRange : array_like (n, m)
        Range in meters
    slantRangeRate : array_like (n, m)
        Slant range rate in meters per second.
    """
    import warnings
    warnings.warn("This function is deprecated; use "
                  "ssapy.compute.radec(..., rate=True)")
    ra, dec, slant, raRate, decRate, slantRate = radec(
        orbit, time, obsPos=obsPos, obsVel=obsVel, observer=observer,
        propagator=propagator, obsAngleCorrection=obsAngleCorrection
    )

    return raRate, decRate, slantRate


def rvObsToRaDecRate(r, v, obsPos=None, obsVel=None):
    """Convert object and observer position and velocity to angles.

    This only does the geometric part; it ignores light travel time,
    which may be applied to the object location before input.  Assumes
    that r, v, obsPos, and obsVel all have common shape.

    Parameters
    ----------
    r : array_like (..., 3)
        object position in meters
    v : array_like (..., 3)
        object velocity in meters per second
    obsPos : array_like (..., 3)
        observer position in meters
    obsVel : array_like (..., 3), optional
        observer velocity in meters per second

    Returns
    -------
    ra : array_like (...)
        right ascension in radians
    dec : array_link (...)
        declination in radians
    slantRange : array_like (...)
        Range in meters
    raRate : array_like (...)
        Rate of change of right ascension*cos(dec) in radians per second.
    decRate : array_like (...)
        Rate of change of declination in radians per second.
    slantRangeRate : array_like (...)
        Slant range rate in meters per second.
    """
    if obsPos is None:
        obsPos = np.zeros_like(r)

    dr = r - obsPos
    slantRange = norm(dr)
    ra = np.arctan2(dr[..., 1], dr[..., 0])
    dec = np.arcsin(dr[..., 2] / slantRange)

    if obsVel is None:
        obsVel = np.zeros_like(v)
    dv = v - obsVel

    # Now need to rotate to ra/dec coords
    sd, cd = np.sin(dec), np.cos(dec)
    sa, ca = np.sin(ra), np.cos(ra)

    rHat = normed(dr)  # (n, m, 3)

    raHat = np.zeros_like(rHat, dtype=float)
    raHat[..., 0] = -sa
    raHat[..., 1] = ca

    decHat = np.zeros_like(raHat)
    decHat[..., 0] = -sd * ca
    decHat[..., 1] = -sd * sa
    decHat[..., 2] = cd

    raRate = np.einsum("...i,...i->...", dv, raHat) / slantRange
    decRate = np.einsum("...i,...i->...", dv, decHat) / slantRange
    rangeRate = np.einsum("...i,...i->...", dv, rHat)
    return ra, dec, slantRange, raRate, decRate, rangeRate


def radecRateObsToRV(ra, dec, slantRange, raRate=None, decRate=None, slantRangeRate=None, obsPos=None, obsVel=None):
    """Convert object angles and observer position to 3D observer position

    This only does the geometric part; it ignores light travel time.  This
    is the inverse of rvObsToRaDecRate.

    If obsVel is None, then the returned velocity will also be None.

    Parameters
    ----------
    ra : array_like (...)
        right ascension in radians
    dec : array_like (...)
        declination in radians
    slantRange : array_like (...)
        Range in meters
    raRate : array_like (...)
        Rate of change of right ascension*cos(dec) in radians per second.
    decRate : array_like (...)
        Rate of change of declination in radians per second.
    slantRangeRate : array_like (...)
        Slant range rate in meters per second.
    obsPos : array_like (..., 3)
        Observer position in meters
    obsVel : array_like (..., 3)
        Observer velocity in meters

    Returns
    -------
    r : array_like (..., 3)
        object position in meters
    v : array_like (..., 3)
        object velocity in meters per second
        observer velocity in meters per second

    v is None if obsVel is None.
    """

    if obsPos is None:
        return ValueError('obsPos must be set!')

    rHat = lb_to_unit(ra, dec)

    sd, cd = np.sin(dec), np.cos(dec)
    sa, ca = np.sin(ra), np.cos(ra)

    raHat = np.zeros_like(rHat, dtype=float)
    raHat[..., 0] = -sa
    raHat[..., 1] = ca

    decHat = np.zeros_like(raHat)
    decHat[..., 0] = -sd * ca
    decHat[..., 1] = -sd * sa
    decHat[..., 2] = cd
    if len(rHat.shape) > 1:
        slantRange = np.array(slantRange)[..., None]
        raRate = np.array(raRate)[..., None]
        decRate = np.array(decRate)[..., None]
        slantRangeRate = np.array(slantRangeRate)[..., None]

    r = obsPos + rHat * slantRange
    if obsVel is None:
        v = None
    else:
        v = (obsVel + rHat * slantRangeRate + slantRange * (raHat * raRate + decHat * decRate))
    return r, v


def earthShadowCoords(r, time):
    """Determine components of position `r` parallel and perpendicular to sun
    unit vector.

    The sun unit vector points from the center of the sun through the center of
    the Earth.  Decomposing a satellite position into these coordinates yields a
    simple model of whether or not the satellite is in Earth's shadow:

        r_par, r_perp = earthShadowCoords(r, time)
        inShadow = r_par > 0 and r_perp < EARTH_RADIUS

    Parameters
    ----------
    r : ndarray (3,)
        Position (GCRF) in meters
    time : float or astropy.time.Time
        If float, then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC

    Returns
    -------
    r_par : float
        Position of satellite projected onto the sun unit vector in meters.
    r_perp : float
        Distance of satellite from Earth-Sun line in meters.
    """
    if isinstance(time, Time):
        time = time.gps
    r_sun = sunPos(time, fast=False)
    if r.ndim == 1:
        n_sun = - r_sun / norm(r_sun)
        r_par = np.dot(r, n_sun)
        r_perp = norm(r - n_sun * r_par)
    else:
        n_sun = - r_sun / norm(r_sun)[:, None]
        r_par = np.sum(r * n_sun, axis=1)
        r_perp = norm(r - n_sun * r_par[:, None])
    return r_par, r_perp


def find_passes(
    orbit, observers,
    tStart, tSpan, dt,
    propagator=KeplerianPropagator(),
    horizon=np.deg2rad(20)
):
    """Find satellite overhead passes for a collection of observers.

    Uses a brute force test of a grid of time points from tStart to tStart+tSpan
    separated by dt.

    Returns passes even if they occur during the daytime or if the satellite is
    not illuminated by the sun.  The only criterion for a successful "pass" is
    for the topocentric altitude of the satellite to be above the input
    `horizon`.  More details about a pass can subsequently be obtained by
    running the `refine_passes` function.

    Note this function is only suitable for `EarthObserver`s and not
    `OrbitalObserver`s.

    Parameters
    ----------
    orbit : Orbit
        Satellite orbit in question.
    observers : List of EarthObserver.
        Earth observers for which to check satellite visibility.
    tStart : float or astropy.time.Time
        Beginning of search window.
        If float, then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    tSpan : float or Quantity
        Time span in which to search for passes.
        If float, then seconds.
    dt : float or Quantity
        Time increment to use during search.  Satellite visibility will be
        computed every dt increment.  Smaller values will decrease the
        probability that a short duration pass will be missed, but will make the
        search take longer to complete.
        If float, then seconds.
    propagator : Propagator, optional
        The propagator instance to use.
    horizon : float or Quantity, optional
        Minimum altitude for which to consider a satellite "visible".
        If float, then should be in radians.

    Returns
    -------
    passDict : dict
        keys are `EarthObserver`s.
        values are lists (possibly empty) of `astropy.Time` corresponding to
        visible passes of the satellite.  Only one time is returned per pass,
        so multiple times in the return list indicate multiple distinct passes.
    """
    import astropy.units as u
    if isinstance(tStart, Time):
        tStart = tStart.gps
    if isinstance(tSpan, u.Quantity):
        tSpan = tSpan.to(u.s).value
    if isinstance(dt, u.Quantity):
        dt = dt.to(u.s).value
    if isinstance(horizon, u.Quantity):
        horizon = horizon.to(u.rad).value
    times = np.arange(tStart, tStart + tSpan, dt)
    out = {}
    for observer in observers:
        out[observer] = []
        alt, _ = quickAltAz(orbit, times, observer, propagator=propagator)
        w = np.nonzero(alt > horizon)[0]
        for wi in w:
            if (wi - 1) in w:  # already caught this pass
                continue
            out[observer].append(Time(times[wi], format='gps'))
    return out


def refine_pass(
    orbit,
    observer,
    time,
    propagator=KeplerianPropagator(),
    horizon=np.deg2rad(20),
    maxSpan=86400.0
):
    """Refine a satellite overhead pass.

    Parameters
    ----------
    orbit : Orbit
        Orbit in question.
    observer : EarthObserver.
        Observer for which to refine pass.
    time : float or astropy.time.Time
        A time when satellite is visible to observer.  (Found using
        find_passes, for instance)
        If float, then should correspond to GPS seconds; i.e., seconds
        since 1980-01-06 00:00:00 UTC
    propagator : Propagator, optional
        The propagator instance to use.
    horizon : float or Quantity, optional
        Minimum altitude for which to consider a satellite "visible".
        If float, then should be in radians.
    maxSpan : float or Quantity, optional
        Maximum amount of time before or after `time` to search for
        rise/set times, or time of max altitude.
        If float, then seconds.

    Returns
    -------
    dict
        Key/values are:
        tStart : Time
            Time at which Orbit rises above horizon
        tEnd : Time
            Time at which Orbit rises above horizon
        tMaxAlt : Time
            Time Orbit passes through maximum altitude
        maxAlt : Quantity
            Maximum altitude
        duration : Quantity
            Duration of pass.
        illumAtStart : bool
            Is satellite illuminated at `tStart`?
        illumAtEnd : bool
            Is satellite illuminated at `tEnd`?
        tTerminator : Time or None
            If illumAtStart != illumAtEnd, then time satellite passes through
            cylindrical terminator shadow.  Otherwise, None.
        sunAltStart : Quantity
            Altitude of sun at `tStart`
        sunAltEnd : Quantity
            Altitude of sun at `tEnd`
    """
    from scipy.optimize import bisect, minimize_scalar
    import astropy.units as u
    import warnings

    if isinstance(time, Time):
        time = time.gps
    if isinstance(horizon, u.Quantity):
        horizon = horizon.to(u.rad).value
    if isinstance(maxSpan, u.Quantity):
        maxSpan = maxSpan.to(u.s).value

    def dalt(t):
        return (
            horizon - quickAltAz(orbit, t, observer, propagator=propagator)[0]
        )

    # Height "outside" of cylindrical Earth shadow model.
    def dshadow(t):
        r, _ = rv(orbit, t, propagator=propagator)
        r_par, r_perp = earthShadowCoords(r, t)
        height = r_perp - EARTH_RADIUS
        if r_par < 0:
            # a bit of a hack, if the sat is on the sun side, then we'll
            # return abs(dshadow) so that the value is positive, which indicates
            # illumination.
            return abs(height)
        else:
            return height

    # Bracket and then bisect to find rise/set times
    def bracket(f, x, dx, xmax):
        x0, x1 = x, x + dx
        f0, f1 = f(x0), f(x1)
        while f0 * f1 > 0 and abs(x1 - x0) < xmax:
            x1 += dx
            f1 = f(x1)
        return x1

    tLow = bracket(dalt, time, -300, maxSpan)
    if dalt(tLow) * dalt(time) > 0:  # didn't bracket, use tLow and issue warning
        warnings.warn("Failed to bracket.  tStart is not rise time!")
        tStart = tLow
    else:
        tStart = bisect(dalt, tLow, time)
    tStart = Time(tStart, format='gps')
    tStart.format = 'iso'

    tHigh = bracket(dalt, time, 300, maxSpan)
    if dalt(tHigh) * dalt(time) > 0:
        warnings.warn("Failed to bracket.  tEnd is not set time!")
        tEnd = tHigh
    else:
        tEnd = bisect(dalt, time, tHigh)
    tEnd = Time(tEnd, format='gps')
    tEnd.format = 'iso'

    # Find maximum altitude
    dalt1 = dalt(tStart.gps)
    dalt2 = dalt(time)
    dalt3 = dalt(tEnd.gps)
    if dalt2 < dalt1 and dalt2 < dalt3:
        result = minimize_scalar(dalt, (tStart.gps, time, tEnd.gps))
        tMaxAlt = Time(result.x, format='gps')
        tMaxAlt.format = 'iso'
        maxAlt = -result.fun + horizon
    else:
        ama = np.argmin([dalt1, dalt2, dalt3])
        tMaxAlt = Time([tStart.gps, time, tEnd.gps][ama], format='gps')
        tMaxAlt.format = 'iso'
        maxAlt = -[dalt1, dalt2, dalt3][ama] + horizon
    duration = (tEnd - tStart).to(u.min)

    # Find illumination state
    illumAtStart = dshadow(tStart.gps) > 0
    sunAltStart = np.rad2deg(observer.sunAlt(tStart.gps)) * u.deg
    illumAtEnd = dshadow(tEnd.gps) > 0
    sunAltEnd = np.rad2deg(observer.sunAlt(tEnd.gps)) * u.deg

    # If we transition between shadow/illuminated, then find the time of
    # transition (time of terminator pass)
    if illumAtStart != illumAtEnd:
        tTerminator = bisect(dshadow, tStart.gps, tEnd.gps)
        tTerminator = Time(tTerminator, format='gps')
        tTerminator.format = 'iso'
    else:
        tTerminator = None

    return {
        "orbit": orbit,
        "observer": observer,
        "propagator": propagator,
        "horizon": horizon,
        "tStart": tStart,
        "tEnd": tEnd,
        "altStart": np.rad2deg(-dalt(tStart.gps) + horizon) * u.deg,
        "altEnd": np.rad2deg(-dalt(tEnd.gps) + horizon) * u.deg,
        "tMaxAlt": tMaxAlt,
        "maxAlt": np.rad2deg(maxAlt) * u.deg,
        "duration": duration,
        "illumAtStart": illumAtStart,
        "illumAtEnd": illumAtEnd,
        "tTerminator": tTerminator,
        "sunAltStart": sunAltStart,
        "sunAltEnd": sunAltEnd
    }


def nby3shape(arr_):
    """
    Reshapes or transforms an input NumPy array into a shape compatible with (n, 3).

    This function takes a NumPy array and ensures it has a shape of `(n, 3)`, 
    where `n` is the number of rows. The behavior depends on the dimensionality of the input array:

    - If the input array is 1-dimensional, it reshapes it into a `(1, 3)` array.
    - If the input array is 2-dimensional:
        - If the second dimension already has size 3, the array is returned as-is.
        - Otherwise, the array is transposed.

    Parameters:
        arr_ (numpy.ndarray): The input NumPy array to reshape or transform.

    Returns:
        numpy.ndarray: The reshaped or transformed array with a shape of `(n, 3)`.

    Raises:
        ValueError: If the input array cannot be reshaped or transformed into the desired shape.

    Examples:
        >>> import numpy as np
        >>> arr1 = np.array([1, 2, 3])
        >>> nby3shape(arr1)
        array([[1, 2, 3]])

        >>> arr2 = np.array([[1, 2, 3], [4, 5, 6]])
        >>> nby3shape(arr2)
        array([[1, 2, 3],
               [4, 5, 6]])

        >>> arr3 = np.array([[1, 4], [2, 5], [3, 6]])
        >>> nby3shape(arr3)
        array([[1, 2, 3],
               [4, 5, 6]])
    """

    if arr_.ndim == 1:
        return np.reshape(arr_, (1, 3))
    if arr_.ndim == 2:
        if np.shape(arr_)[1] == 3:
            return arr_
        else:
            return arr_.T


def calculate_orbital_elements(r_, v_, mu_barycenter=EARTH_MU):
    """
    Calculate the Keplerian orbital elements from position and velocity vectors.

    This function computes the Keplerian orbital elements (semi-major axis, eccentricity, inclination, true longitude, argument of periapsis, longitude of ascending node, true anomaly, and specific angular momentum) for one or more celestial objects given their position and velocity vectors.

    Parameters:
    ----------
    r_ : (n, 3) numpy.ndarray
        Array of position vectors (in meters) of the celestial objects. Each row represents a position vector.
    v_ : (n, 3) numpy.ndarray
        Array of velocity vectors (in meters per second) of the celestial objects. Each row represents a velocity vector.
    mu_barycenter : float, optional
        Gravitational parameter (standard gravitational constant times mass) of the central body (default is `EARTH_MU`). This parameter defines the gravitational influence of the central body on the orbiting object.

    Returns:
    -------
    dict
        A dictionary containing the orbital elements for each celestial object:
        - 'a': Semi-major axis (in meters).
        - 'e': Eccentricity (dimensionless).
        - 'i': Inclination (in radians).
        - 'tl': True longitude (in radians).
        - 'ap': Argument of periapsis (in radians).
        - 'raan': Longitude of ascending node (in radians).
        - 'ta': True anomaly (in radians).
        - 'L': Specific angular momentum (in meters squared per second).

    Notes:
    ------
    - Position and velocity vectors should be provided in the same units.
    - The function assumes that the input vectors are provided in an array where each row corresponds to a different celestial object.
    - Orbital elements are computed using standard orbital mechanics formulas.
    - The inclination is measured from the reference plane, and the argument of periapsis and true anomaly are measured in the orbital plane.

    Example:
    --------
    >>> r = np.array([[1e7, 1e7, 1e7], [1e8, 1e8, 1e8]])
    >>> v = np.array([[1e3, 2e3, 3e3], [4e3, 5e3, 6e3]])
    >>> calculate_orbital_elements(r, v, mu_barycenter=3.986e14)
    {'a': [1.5707e7, 2.234e8], 'e': [0.123, 0.456], 'i': [0.785, 0.654], 'tl': [2.345, 3.456], 'ap': [0.123, 0.456], 'raan': [1.234, 2.345], 'ta': [0.567, 1.678], 'L': [1.234e10, 2.345e11]}
    """
    # mu_barycenter - all bodies interior to Earth
    # 1.0013415732186798 #All bodies of solar system
    mu_ = mu_barycenter
    rarr = nby3shape(r_)
    varr = nby3shape(v_)
    aarr = []
    earr = []
    incarr = []
    true_longitudearr = []
    argument_of_periapsisarr = []
    longitude_of_ascending_nodearr = []
    true_anomalyarr = []
    hmagarr = []
    for r, v in zip(rarr, varr):
        r = np.array(r)  # print(f'r: {r}')
        v = np.array(v)  # print(f'v: {v}')

        rmag = np.sqrt(r.dot(r))
        vmag = np.sqrt(v.dot(v))

        h = np.cross(r, v)
        hmag = np.sqrt(h.dot(h))
        n = np.cross(np.array([0, 0, 1]), h)

        a = 1 / ((2 / rmag) - (vmag ** 2) / mu_)

        evector = np.cross(v, h) / (mu_) - r / rmag
        e = np.sqrt(evector.dot(evector))

        inc = np.arccos(h[2] / hmag)

        if np.dot(r, v) > 0:
            true_anomaly = np.arccos(np.dot(evector, r) / (e * rmag))
        else:
            true_anomaly = 2 * np.pi - np.arccos(np.dot(evector, r) / (e * rmag))
        if evector[2] >= 0:
            argument_of_periapsis = np.arccos(np.dot(n, evector) / (e * np.sqrt(n.dot(n))))
        else:
            argument_of_periapsis = 2 * np.pi - np.arccos(np.dot(n, evector) / (e * np.sqrt(n.dot(n))))
        if n[1] >= 0:
            longitude_of_ascending_node = np.arccos(n[0] / np.sqrt(n.dot(n)))
        else:
            longitude_of_ascending_node = 2 * np.pi - np.arccos(n[0] / np.sqrt(n.dot(n)))

        true_longitude = true_anomaly + argument_of_periapsis + longitude_of_ascending_node
        aarr.append(a)
        earr.append(e)
        incarr.append(inc)
        true_longitudearr.append(true_longitude)
        argument_of_periapsisarr.append(argument_of_periapsis)
        longitude_of_ascending_nodearr.append(longitude_of_ascending_node)
        true_anomalyarr.append(true_anomaly)
        hmagarr.append(hmag)
    return {
        'a': aarr,
        'e': earr,
        'i': incarr,
        'tl': true_longitudearr,
        'ap': argument_of_periapsisarr,
        'raan': longitude_of_ascending_nodearr,
        'ta': true_anomalyarr,
        'L': hmagarr
    }


######################################################################################
# Lambertian brightness functions
######################################################################################


def moon_shine(r_moon, r_sat, r_earth, r_sun, radius, albedo, albedo_moon, albedo_back, albedo_front, area_panels):  # In SI units, takes single values or arrays returns a fractional flux
    """
    Calculate the fractional flux of sunlight reflected from the Moon to the satellite.

    This function computes the flux of sunlight reflected from the Moon to the satellite, including contributions from both the front and back surfaces of the satellite's solar panels.

    Parameters:
    ----------
    r_moon : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Moon.
    r_sat : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the satellite.
    r_earth : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Earth.
    r_sun : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Sun.
    radius : float
        Radius of the satellite in meters.
    albedo : float
        Albedo of the satellite's surface.
    albedo_moon : float
        Albedo of the Moon.
    albedo_back : float
        Albedo of the back surface of the satellite's solar panels.
    albedo_front : float
        Albedo of the front surface of the satellite's solar panels.
    area_panels : float
        Area of the satellite's solar panels in square meters.

    Returns:
    -------
    dict
        Dictionary containing the flux contributions from the Moon to the satellite:
        - 'moon_bus': Fraction of light reflected off the satellite's bus from the Moon.
        - 'moon_panels': Fraction of light reflected off the satellite's panels from the Moon.

    Notes:
    ------
    - The function assumes that the solar panels are always facing the Sun and calculates flux based on the phase angles.
    - Flux contributions from both the front and back surfaces of the solar panels are computed.

    Example:
    --------
    >>> r_moon = np.array([[1e8, 1e8, 1e8]])
    >>> r_sat = np.array([[1e7, 1e7, 1e7]])
    >>> r_earth = np.array([[0, 0, 0]])
    >>> r_sun = np.array([[1e11, 0, 0]])
    >>> moon_shine(r_moon, r_sat, r_earth, r_sun, radius=0.4, albedo=0.20, albedo_moon=0.12, albedo_back=0.50, albedo_front=0.05, area_panels=100)
    {'moon_bus': array([...]), 'moon_panels': array([...])}
    """
    # https://amostech.com/TechnicalPapers/2013/POSTER/COGNION.pdf
    moon_phase_angle = get_angle(r_sun, r_moon, r_sat)  # Phase of the moon as viewed from the sat.
    sun_angle = get_angle(r_sun, r_sat, r_moon)  # angle from Sun to object to Earth
    moon_to_earth_angle = get_angle(r_moon, r_sat, r_earth)
    r_moon_sat = np.linalg.norm(r_sat - r_moon, axis=-1)
    r_earth_sat = np.linalg.norm(r_sat - r_earth, axis=-1)  # Earth is the observer.
    flux_moon_to_sat = 2 / 3 * albedo_moon * MOON_RADIUS**2 / (np.pi * (r_moon_sat)**2) * (np.sin(moon_phase_angle) + (np.pi - moon_phase_angle) * np.cos(moon_phase_angle))  # Fraction of sunlight reflected from the Moon to satellite
    # Fraction of light from back of solar panel
    flux_back = np.zeros_like(sun_angle)
    flux_back[sun_angle > np.pi / 2] = np.abs(albedo_back * area_panels / (np.pi * r_earth_sat[sun_angle > np.pi / 2]**2) * np.cos(np.pi - moon_to_earth_angle[sun_angle > np.pi / 2]) * flux_moon_to_sat[sun_angle > np.pi / 2])  # Fraction of Moon light reflected off back of solar panels - which are assumed to be always facing the Sun. Angle: Sun - Observer - Sat
    flux_front = np.zeros_like(sun_angle)
    flux_front[sun_angle < np.pi / 2] = np.abs(albedo_front * area_panels / (np.pi * r_earth_sat[sun_angle < np.pi / 2]**2) * np.cos(moon_to_earth_angle[sun_angle < np.pi / 2]) * flux_moon_to_sat[sun_angle < np.pi / 2])  # Fraction of Sun light scattered off front of the solar panels - which are assumed to be always facing the Sun. Angle: Sun - Sat - Observer
    flux_panels = flux_back + flux_front
    flux_bus = 2 / 3 * albedo * radius**2 / (np.pi * r_earth_sat**2) * flux_moon_to_sat
    return {'moon_bus': flux_bus, 'moon_panels': flux_panels}


def earth_shine(r_sat, r_earth, r_sun, radius, albedo, albedo_earth, albedo_back, area_panels):  # In SI units, takes single values or arrays returns a flux
    """
    Calculate the fractional flux of sunlight reflected from the Earth to the satellite.

    This function computes the flux of sunlight reflected from the Earth to the satellite, including contributions from the back surface of the satellite's solar panels.

    Parameters:
    ----------
    r_sat : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the satellite.
    r_earth : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Earth.
    r_sun : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Sun.
    radius : float
        Radius of the satellite in meters.
    albedo : float
        Albedo of the satellite's surface.
    albedo_earth : float
        Albedo of the Earth.
    albedo_back : float
        Albedo of the back surface of the satellite's solar panels.
    area_panels : float
        Area of the satellite's solar panels in square meters.

    Returns:
    -------
    dict
        Dictionary containing the flux contributions from the Earth to the satellite:
        - 'earth_bus': Fraction of light reflected off the satellite's bus from the Earth.
        - 'earth_panels': Fraction of light reflected off the satellite's panels from the Earth.

    Notes:
    ------
    - The function assumes that the solar panels are always facing the Sun and calculates flux based on the phase angle.
    - Flux contributions from the back surface of the solar panels are computed.

    Example:
    --------
    >>> r_sat = np.array([[1e7, 1e7, 1e7]])
    >>> r_earth = np.array([[1.496e11, 0, 0]])
    >>> r_sun = np.array([[0, 0, 0]])
    >>> earth_shine(r_sat, r_earth, r_sun, radius=0.4, albedo=0.20, albedo_earth=0.30, albedo_back=0.50, area_panels=100)
    {'earth_bus': array([...]), 'earth_panels': array([...])}
    """
    # https://amostech.com/TechnicalPapers/2013/POSTER/COGNION.pdf
    phase_angle = get_angle(r_sun, r_sat, r_earth)  # angle from Sun to object to Earth
    earth_angle = np.pi - phase_angle  # Sun to Earth to object.
    r_earth_sat = np.linalg.norm(r_sat - r_earth, axis=-1)  # Earth is the observer.
    flux_earth_to_sat = 2 / 3 * albedo_earth * EARTH_RADIUS**2 / (np.pi * (r_earth_sat)**2) * (np.sin(earth_angle) + (np.pi - earth_angle) * np.cos(earth_angle))  # Fraction of sunlight reflected from the Earth to satellite
    # Fraction of light from back of solar panel
    flux_back = np.zeros_like(phase_angle)
    flux_back[phase_angle > np.pi / 2] = albedo_back * area_panels / (np.pi * r_earth_sat[phase_angle > np.pi / 2]**2) * np.cos(np.pi - phase_angle[phase_angle > np.pi / 2]) * flux_earth_to_sat[phase_angle > np.pi / 2]  # Fraction of Earth light reflected off back of solar panels - which are assumed to be always facing the Sun. Angle: Sun - Observer - Sat
    flux_bus = 2 / 3 * albedo * radius**2 / (np.pi * r_earth_sat**2) * flux_earth_to_sat
    return {'earth_bus': flux_bus, 'earth_panels': flux_back}


def sun_shine(r_sat, r_earth, r_sun, radius, albedo, albedo_front, area_panels):  # In SI units, takes single values or arrays returns a fractional flux
    """
    Calculate the fractional flux of sunlight reflected from the Sun to the satellite.

    This function computes the flux of sunlight reflected from the Sun to the satellite, including contributions from the front surface of the satellite's solar panels.

    Parameters:
    ----------
    r_sat : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the satellite.
    r_earth : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Earth.
    r_sun : (n, 3) numpy.ndarray
        Array of coordinates representing the position of the Sun.
    radius : float
        Radius of the satellite in meters.
    albedo : float
        Albedo of the satellite's surface.
    albedo_front : float
        Albedo of the front surface of the satellite's solar panels.
    area_panels : float
        Area of the satellite's solar panels in square meters.

    Returns:
    -------
    dict
        Dictionary containing the flux contributions from the Sun to the satellite:
        - 'sun_bus': Fraction of light reflected off the satellite's bus from the Sun.
        - 'sun_panels': Fraction of light reflected off the satellite's panels from the Sun.

    Notes:
    ------
    - The function assumes that the solar panels are always facing the Sun and calculates flux based on the phase angle.
    - Flux contributions from the front surface of the solar panels are computed.

    Example:
    --------
    >>> r_sat = np.array([[1e7, 1e7, 1e7]])
    >>> r_earth = np.array([[0, 0, 0]])
    >>> r_sun = np.array([[1e11, 0, 0]])
    >>> sun_shine(r_sat, r_earth, r_sun, radius=0.4, albedo=0.20, albedo_front=0.05, area_panels=100)
    {'sun_bus': array([...]), 'sun_panels': array([...])}
    """
    # https://amostech.com/TechnicalPapers/2013/POSTER/COGNION.pdf
    phase_angle = get_angle(r_sun, r_sat, r_earth)  # angle from Sun to object to Earth
    r_earth_sat = np.linalg.norm(r_sat - r_earth, axis=-1)  # Earth is the observer.
    flux_front = np.zeros_like(phase_angle)
    flux_front[phase_angle < np.pi / 2] = albedo_front * area_panels / (np.pi * r_earth_sat[phase_angle < np.pi / 2]**2) * np.cos(phase_angle[phase_angle < np.pi / 2])  # Fraction of Sun light scattered off front of the solar panels - which are assumed to be always facing the Sun. Angle: Sun - Sat - Observer
    flux_bus = 2 / 3 * albedo * radius**2 / (np.pi * (r_earth_sat)**2) * (np.sin(phase_angle) + (np.pi - phase_angle) * np.cos(phase_angle))  # Fraction of light reflected off satellite from Sun
    return {'sun_bus': flux_bus, 'sun_panels': flux_front}


def calc_M_v(r_sat, r_earth, r_sun, r_moon=False, radius=0.4, albedo=0.20, sun_Mag=4.80, albedo_earth=0.30, albedo_moon=0.12, albedo_back=0.50, albedo_front=0.05, area_panels=100, return_components=False):
    """
    Calculate the apparent magnitude (M_v) of a satellite due to reflections from the Sun, Earth, and optionally the Moon.

    This function computes the apparent magnitude of a satellite based on its reflected light from the Sun, Earth, and optionally the Moon. It uses separate functions to calculate the flux contributions from each of these sources and combines them to determine the overall apparent magnitude.

    Parameters:
    ----------
    r_sat : (n, 3) numpy.ndarray
        Position of the satellite in meters.
    r_earth : (n, 3) numpy.ndarray
        Position of the Earth in meters.
    r_sun : (n, 3) numpy.ndarray
        Position of the Sun in meters.
    r_moon : (n, 3) numpy.ndarray or False, optional
        Position of the Moon in meters. If False, the Moon's contribution is ignored (default is False).
    radius : float, optional
        Radius of the satellite in meters (default is 0.4 m).
    albedo : float, optional
        Albedo of the satellite's surface, representing its reflectivity (default is 0.20).
    sun_Mag : float, optional
        Solar magnitude (apparent magnitude of the Sun) used in magnitude calculations (default is 4.80).
    albedo_earth : float, optional
        Albedo of the Earth, representing its reflectivity (default is 0.30).
    albedo_moon : float, optional
        Albedo of the Moon, representing its reflectivity (default is 0.12).
    albedo_back : float, optional
        Albedo of the back surface of the satellite (default is 0.50).
    albedo_front : float, optional
        Albedo of the front surface of the satellite (default is 0.05).
    area_panels : float, optional
        Area of the satellite's panels in square meters (default is 100 m^2).
    return_components : bool, optional
        If True, returns the magnitude as well as the flux components from the Sun, Earth, and Moon (default is False).

    Returns:
    -------
    float
        The apparent magnitude (M_v) of the satellite as observed from Earth.

    dict, optional
        If `return_components` is True, a dictionary containing the flux components from the Sun, Earth, and Moon.

    Notes:
    ------
    - The function uses separate calculations for flux contributions from the Sun, Earth, and Moon:
        - `sun_shine` calculates the flux from the Sun.
        - `earth_shine` calculates the flux from the Earth.
        - `moon_shine` calculates the flux from the Moon (if applicable).
    - The apparent magnitude is calculated based on the distances between the satellite, Sun, Earth, and optionally the Moon, as well as their respective albedos and other parameters.

    Example usage:
    --------------
    >>> r_sat = np.array([[1e7, 2e7, 3e7]])
    >>> r_earth = np.array([1.496e11, 0, 0])
    >>> r_sun = np.array([0, 0, 0])
    >>> Mag_v = calc_M_v(r_sat, r_earth, r_sun, return_components=True)
    >>> Mag_v
    (15.63, {'sun_bus': 0.1, 'sun_panels': 0.2, 'earth_bus': 0.05, 'earth_panels': 0.1, 'moon_bus': 0.03, 'moon_panels': 0.07})
    """
    r_sun_sat = np.linalg.norm(r_sat - r_sun, axis=-1)
    frac_flux_sun = {'sun_bus': 0, 'sun_panels': 0}
    frac_flux_earth = {'earth_bus': 0, 'earth_panels': 0}
    frac_flux_moon = {'moon_bus': 0, 'moon_panels': 0}
    frac_flux_sun = sun_shine(r_sat, r_earth, r_sun, radius, albedo, albedo_front, area_panels)
    frac_flux_earth = earth_shine(r_sat, r_earth, r_sun, radius, albedo, albedo_earth, albedo_back, area_panels)
    if r_moon is not False:
        frac_flux_moon = moon_shine(r_moon, r_sat, r_earth, r_sun, radius, albedo, albedo_moon, albedo_back, albedo_front, area_panels)
    merged_dict = {**frac_flux_sun, **frac_flux_earth, **frac_flux_moon}
    total_frac_flux = sum(merged_dict.values())
    Mag_v = (2.5 * np.log10((r_sun_sat / (10 * u.Unit('parsec').to(u.Unit('m'))))**2) + sun_Mag) - 2.5 * np.log10(total_frac_flux)
    if return_components:
        return Mag_v, merged_dict
    else:
        return Mag_v


def M_v_lambertian(r_sat, times, radius=1.0, albedo=0.20, sun_Mag=4.80, albedo_earth=0.30, albedo_moon=0.12, plot=False):
    """
    Calculate the apparent magnitude (M_v) of a satellite due to reflections from the Sun, Earth, and Moon.

    This function computes the apparent magnitude of a satellite based on its reflected light from the Sun, Earth, and Moon, using the Lambertian reflection model. It optionally generates plots to visualize the results.

    Parameters:
    ----------
    r_sat : (n, 3) numpy.ndarray
        Position of the satellite in meters.
    times : Time or array_like
        The times corresponding to the satellite and celestial body positions. Used to obtain the positions of the Sun and Moon.
    radius : float, optional
        Radius of the satellite in meters (default is 1.0 m).
    albedo : float, optional
        Albedo of the satellite's surface, representing its reflectivity (default is 0.20).
    sun_Mag : float, optional
        Solar magnitude (apparent magnitude of the Sun) used in magnitude calculations (default is 4.80).
    albedo_earth : float, optional
        Albedo of the Earth, representing its reflectivity (default is 0.30).
    albedo_moon : float, optional
        Albedo of the Moon, representing its reflectivity (default is 0.12).
    plot : bool, optional
        If True, generates plots to visualize solar phase angle and magnitudes (default is False).

    Returns:
    -------
    numpy.ndarray
        The apparent magnitude (M_v) of the satellite as observed from Earth, considering reflections from the Sun, Earth, and Moon.

    Notes:
    ------
    - The function uses a Lambertian reflection model to compute the fraction of sunlight reflected by the satellite, Earth, and Moon.
    - The apparent magnitude is calculated based on the distances between the satellite, Sun, Earth, and Moon, as well as their respective albedos.
    - The function generates four subplots if `plot` is set to True:
      1. Solar phase angle of the satellite.
      2. Solar magnitude (M_v) of the satellite due to the Sun.
      3. Magnitude (M_v) of the satellite due to reflections from the Earth.
      4. Magnitude (M_v) of the satellite due to reflections from the Moon.

    Example usage:
    --------------
    >>> r_sat = np.array([[1e7, 2e7, 3e7]])
    >>> times = Time("2024-01-01")
    >>> M_v = M_v_lambertian(r_sat, times, plot=True)
    >>> M_v
    array([15.63])
    """
    pc_to_m = 3.085677581491367e+16
    r_sun = get_body('Sun').position(times).T
    r_moon = get_body('Moon').position(times).T
    r_earth = np.zeros_like(r_sun)

    r_sun_sat = np.linalg.norm(r_sat - r_sun, axis=-1)
    r_earth_sat = np.linalg.norm(r_sat, axis=-1)
    r_moon_sat = np.linalg.norm(r_sat - r_moon, axis=-1)

    sun_angle = get_angle(r_sun, r_sat, r_earth)
    earth_angle = np.pi - sun_angle
    moon_phase_angle = get_angle(r_sun, r_moon, r_sat)  # Phase of the moon as viewed from the sat.
    moon_to_earth_angle = get_angle(r_moon, r_sat, r_earth)

    flux_moon_to_sat = 2 / 3 * albedo_moon * MOON_RADIUS**2 / (np.pi * (r_moon_sat)**2) * (np.sin(moon_phase_angle) + (np.pi - moon_phase_angle) * np.cos(moon_phase_angle))  # Fraction of sunlight reflected from the Moon to satellite
    flux_earth_to_sat = 2 / 3 * albedo_earth * EARTH_RADIUS**2 / (np.pi * (r_earth_sat)**2) * (np.sin(earth_angle) + (np.pi - earth_angle) * np.cos(earth_angle))  # Fraction of sunlight reflected from the Earth to satellite

    frac_flux_sun = 2 / 3 * albedo * radius**2 / (np.pi * (r_earth_sat)**2) * (np.sin(sun_angle) + (np.pi - sun_angle) * np.cos(sun_angle))  # Fraction of light reflected off satellite from Sun
    frac_flux_earth = 2 / 3 * albedo * radius**2 / (np.pi * r_earth_sat**2) * flux_earth_to_sat
    frac_flux_moon = 2 / 3 * albedo * radius**2 / (np.pi * r_earth_sat**2) * flux_moon_to_sat
    Mag_v = (2.5 * np.log10((r_sun_sat / (10 * pc_to_m))**2) + sun_Mag) - 2.5 * np.log10(frac_flux_sun + frac_flux_earth + frac_flux_moon)
    if plot:
        import matplotlib.pyplot as plt
        sun_scale = 149597870700.0 * (RGEO / np.max(r_earth_sat) ) * 0.75
        color_map ='inferno_r'
        fig = plt.figure(figsize=(18, 4))
        ax = fig.add_subplot(1, 4, 1)
        ax.scatter(r_earth[:, 0], r_earth[:, 1], c='Blue', s=10)
        scatter = ax.scatter(r_sat[:, 0] / RGEO, r_sat[:, 1] / RGEO, c=sun_angle, cmap=color_map)
        colorbar = plt.colorbar(scatter)
        ax.scatter(r_sun[:, 0] / sun_scale, r_sun[:, 1] / sun_scale, c=plt.cm.Oranges(np.linspace(0.25, 0.75, len(r_sat[:, 0]))), s=10)
        ax.set_title('Solar Phase')
        ax.set_xlabel('X [GEO]')
        ax.set_ylabel('Y [GEO]')
        ax.axis('equal')

        ax = fig.add_subplot(1, 4, 2)
        ax.scatter(r_earth[0], r_earth[1], c='Blue', s=10)
        scatter = ax.scatter(r_sat[:, 0] / RGEO, r_sat[:, 1] / RGEO, c=(2.5 * np.log10((r_sun_sat / (10 * pc_to_m))**2) + sun_Mag) - 2.5 * np.log10(frac_flux_sun), cmap=color_map)
        colorbar = plt.colorbar(scatter)
        ax.scatter(r_sun[:, 0] / sun_scale, r_sun[:, 1] / sun_scale, c=plt.cm.Oranges(np.linspace(0.25, 0.75, len(r_sat[:, 0]))), s=10)
        ax.set_title('Solar M_v')
        ax.axis('equal')

        ax = fig.add_subplot(1, 4, 3)
        ax.scatter(r_earth[:, 0], r_earth[:, 1], c='Blue', s=10)
        scatter = ax.scatter(r_sat[:, 0] / RGEO, r_sat[:, 1] / RGEO, c=(2.5 * np.log10((r_sun_sat / (10 * pc_to_m))**2) + sun_Mag) - 2.5 * np.log10(frac_flux_earth), cmap=color_map)
        colorbar = plt.colorbar(scatter)
        ax.scatter(r_sun[:, 0] / sun_scale, r_sun[:, 1] / sun_scale, c=plt.cm.Oranges(np.linspace(0.25, 0.75, len(r_sat[:, 0]))), s=10)

        ax.set_title('Earth M_v')
        ax.axis('equal')

        ax = fig.add_subplot(1, 4, 4)
        ax.scatter(r_earth[:, 0], r_earth[:, 1], c='Blue', s=10)
        scatter = ax.scatter(r_sat[:, 0] / RGEO, r_sat[:, 1] / RGEO, c=(2.5 * np.log10((r_sun_sat / (10 * pc_to_m))**2) + sun_Mag) - 2.5 * np.log10(frac_flux_moon), cmap=color_map)
        ax.scatter(r_moon[:, 0] / RGEO, r_moon[:, 1] / RGEO, c=plt.cm.Greys(np.linspace(0.5, 1, len(r_sat[:, 0]))), s=5)

        colorbar = plt.colorbar(scatter)
        ax.set_title('Lunar M_v')
        ax.axis('equal')
        plt.show()

    return Mag_v


def calc_gamma(r, t):
    """
    Calculate the gamma angle between position and velocity vectors in the ITRF frame.

    Parameters:
    r (numpy.ndarray): The position vectors in the GCRF frame, shaped (n, 3), where n is the number of vectors.
    t (numpy.ndarray or astropy.time.Time): The times corresponding to the position vectors. Can be an array of GPS seconds or an Astropy Time object.

    Returns:
    numpy.ndarray: An array of gamma angles in degrees between the position and velocity vectors for each time point.

    Notes:
    - This function first converts the given position vectors from the GCRF frame to the ITRF frame.
    - It then calculates the angle between the position and velocity vectors at each time point in the ITRF frame.
    - If the input time array is an Astropy Time object, it converts it to GPS time before processing.
    """
    r_itrf, v_itrf = gcrf_to_itrf(r, t, v=True)
    if isinstance(t[0], Time):
        t = t.gps
    gamma = np.degrees(np.apply_along_axis(lambda x: angle_between_vectors(x[:3], x[3:]), 1, np.concatenate((r_itrf, v_itrf), axis=1))) - 90
    return gamma



def moon_normal_vector(t):
    """
    Calculate the normal vector to the Moon's orbital plane at a given time.

    This function computes the normal vector to the Moon's orbital plane by taking the cross product of the Moon's position vectors at two different times: the given time `t` and one week later. The resulting vector is normalized to provide the direction of the orbital plane's normal.

    Parameters:
    ----------
    t : Time
        The time at which to calculate the normal vector to the Moon's orbital plane.

    Returns:
    -------
    numpy.ndarray
        A 3-element array representing the normal vector to the Moon's orbital plane at the given time. The vector is normalized to have a unit length.

    Notes:
    ------
    - The function assumes a circular orbit for the Moon and uses a time step of one week (604800 seconds) to calculate the normal vector.
    - The normal vector is perpendicular to the plane defined by the Moon's position vectors at the given time and one week later.

    Example usage:
    --------------
    >>> t = Time("2024-01-01")
    >>> normal_vector = moon_normal_vector(t)
    >>> normal_vector
    array([-0.093,  0.014,  0.995])
    """
    r = get_body("moon").position(t).T
    r_random = get_body("moon").position(t.gps + 604800).T
    return np.cross(r, r_random) / np.linalg.norm(r, axis=-1)


def lunar_lagrange_points(t):
    """
    Calculate the positions of the lunar Lagrange points in the GCRF frame at a given time.

    This function computes the positions of the five Lagrange points (L1, L2, L3, L4, and L5) in the Earth-Moon system at a specific time `t`. It considers the positions of the Earth and Moon and uses the orbital period of the Moon to find the Lagrange points.

    Parameters:
    ----------
    t : Time
        The time at which to calculate the Lagrange points. The position of the Moon at this time is used to compute the Lagrange points.

    Returns:
    -------
    dict
        A dictionary containing the coordinates of the five Lagrange points:
        - "L1": Position of the first Lagrange point between the Earth and the Moon.
        - "L2": Position of the second Lagrange point beyond the Moon.
        - "L3": Position of the third Lagrange point directly opposite the Moon, relative to the Earth.
        - "L4": Position of the fourth Lagrange point, calculated as the Moon's position offset by one-sixth of the lunar period.
        - "L5": Position of the fifth Lagrange point, calculated as the Moon's position offset by negative one-sixth of the lunar period.

    Notes:
    ------
    - The function assumes a circular orbit for the Moon.
    - The lunar period used is approximately 2.36 million seconds.
    - The gravitational parameters of the Earth and Moon are denoted as `EARTH_MU` and `MOON_MU`, respectively.

    Example usage:
    --------------
    >>> t = Time("2024-01-01")
    >>> lagrange_points = lunar_lagrange_points(t)
    >>> lagrange_points["L1"]
    array([1.02e6, 0.0, 0.0])
    >>> lagrange_points["L4"]
    array([1.5e6, 1.5e6, 0.0])
    """

    r = get_body("moon").position(t).T
    d = np.linalg.norm(r)  # Distance between Earth and Moon
    unit_vector_moon = r / np.linalg.norm(r, axis=-1)
    # plane_vector = np.cross(r, r_random)
    lunar_period_seconds = 2.3605915968e6

    # Coefficients of the quadratic equation
    a = EARTH_MU - MOON_MU
    b = 2 * MOON_MU * d
    c = -MOON_MU * d**2

    # Solve the quadratic equation
    discriminant = b**2 - 4*a*c

    if discriminant >= 0:
        L1_from_moon = (-b - np.sqrt(discriminant)) / (2*a) * unit_vector_moon
        L2_from_moon = (-b + np.sqrt(discriminant)) / (2*a) * unit_vector_moon
    else:
        print("Discriminate is less than 0! THAT'S WEIRD FIX IT.")
        L1_from_moon = None
        L2_from_moon = None

    return {
        "L1": L1_from_moon + r,
        "L2": L2_from_moon + r,
        "L3": -r,
        "L4": get_body("moon").position(t.gps + lunar_period_seconds / 6).T,
        "L5": get_body("moon").position(t.gps - lunar_period_seconds / 6).T
    }


def lunar_lagrange_points_circular(t):
    """
    Calculate the positions of the lunar Lagrange points in the GCRF frame for a given time.

    This function calculates the positions of the five Lagrange points (L1, L2, L3, L4, and L5) in the Earth-Moon system at a specific time `t`. It accounts for the rotation of the Moon's orbit around the Earth, providing the positions in a circular approximation of the Earth-Moon system.

    Parameters:
    ----------
    t : Time
        The time at which to calculate the Lagrange points. The position of the Moon at this time is used to compute the Lagrange points.

    Returns:
    -------
    dict
        A dictionary containing the coordinates of the five Lagrange points:
        - "L1": Position of the first Lagrange point between the Earth and the Moon.
        - "L2": Position of the second Lagrange point beyond the Moon.
        - "L3": Position of the third Lagrange point directly opposite the Moon, relative to the Earth.
        - "L4": Position of the fourth Lagrange point, forming an equilateral triangle with the Earth and Moon.
        - "L5": Position of the fifth Lagrange point, forming an equilateral triangle with the Earth and Moon, but on the opposite side.

    Notes:
    ------
    - The function assumes a circular orbit for the Moon and uses a rotation matrix to align the z-axis with the Moon's normal vector.
    - The positions of L4 and L5 are calculated using a rotation matrix to align with the Moon's orientation.
    - The gravitational parameters of the Earth and Moon are denoted as `EARTH_MU` and `MOON_MU`, respectively.

    Example usage:
    --------------
    >>> t = Time("2024-01-01")
    >>> lagrange_points = lunar_lagrange_points_circular(t)
    >>> lagrange_points["L1"]
    array([1.02e6, 0.0, 0.0])
    >>> lagrange_points["L4"]
    array([1.5e6, 1.5e6, 0.0])
    """
    r = get_body("moon").position(t).T
    d = np.linalg.norm(r)  # Distance between Earth and Moon
    unit_vector_moon = r / np.linalg.norm(r, axis=-1)

    # Coefficients of the quadratic equation
    a = EARTH_MU - MOON_MU
    b = 2 * MOON_MU * d
    c = -MOON_MU * d**2

    # Solve the quadratic equation
    discriminant = b**2 - 4*a*c

    if discriminant >= 0:
        L1_from_moon = (-b - np.sqrt(discriminant)) / (2*a) * unit_vector_moon
        L2_from_moon = (-b + np.sqrt(discriminant)) / (2*a) * unit_vector_moon
    else:
        print("Discriminate is less than 0! THAT'S WEIRD FIX IT.")
        L1_from_moon = None
        L2_from_moon = None

    # L45
    # Create the rotation matrix to align z-axis with the normal vector
    normal_vector = moon_normal_vector(t)
    rotation_matrix = rotation_matrix_from_vectors(np.array([0, 0, 1]), normal_vector)
    theta = np.radians(60) + np.arctan2(unit_vector_moon[1], unit_vector_moon[0])
    L4 = np.vstack((d * np.cos(theta), d * np.sin(theta), np.zeros_like(theta))).T
    L4 = np.squeeze(L4 @ rotation_matrix.T)
    theta = -np.radians(60) + np.arctan2(unit_vector_moon[1], unit_vector_moon[0])
    L5 = np.vstack((d * np.cos(theta), d * np.sin(theta), np.zeros_like(theta))).T
    L5 = np.squeeze(L5 @ rotation_matrix.T)

    return {
        "L1": L1_from_moon + r,
        "L2": L2_from_moon + r,
        "L3": -r,
        "L4": L4,
        "L5": L5
    }


def lagrange_points_lunar_frame():
    """
    Calculate the positions of the lunar Lagrange points in the lunar frame, This frame is defined by the coordinate transformation in utils.py gcrf_to_lunar().

    Returns:
    -------
    dict
        A dictionary containing the coordinates of the five Lagrange points:
        - "L1": Position of the first Lagrange point between the Earth and the Moon.
        - "L2": Position of the second Lagrange point beyond the Moon.
        - "L3": Position of the third Lagrange point directly opposite the Moon, relative to the Earth.
        - "L4": Position of the fourth Lagrange point, forming an equilateral triangle with the Earth and Moon.
        - "L5": Position of the fifth Lagrange point, forming an equilateral triangle with the Earth and Moon, but on the opposite side.

    Notes:
    ------
    - The function assumes that the Earth and Moon are the two primary bodies, with the Earth-Moon distance denoted as `LD`.
    - The gravitational parameters of the Earth and Moon are denoted as `EARTH_MU` and `MOON_MU`, respectively.
    - The positions of L4 and L5 are calculated using the fact that these points form an equilateral triangle with the Earth and Moon.

    Example usage:
    --------------
    >>> lagrange_points = lagrange_points_lunar_frame()
    >>> lagrange_points["L1"]
    array([1.01e6, 0.0, 0.0])
    >>> lagrange_points["L4"]
    array([1.5e6, 1.5e6, 0.0])
    """
    r = np.array([LD / RGEO, 0, 0])
    d = np.linalg.norm(r)  # Distance between Earth and Moon
    unit_vector_moon = r / np.linalg.norm(r, axis=-1)

    # Coefficients of the quadratic equation
    a = EARTH_MU - MOON_MU
    b = 2 * MOON_MU * d
    c = -MOON_MU * d**2

    # Solve the quadratic equation
    discriminant = b**2 - 4*a*c

    if discriminant >= 0:
        L1_from_moon = (-b - np.sqrt(discriminant)) / (2*a) * unit_vector_moon
        L2_from_moon = (-b + np.sqrt(discriminant)) / (2*a) * unit_vector_moon
    else:
        print("Discriminate is less than 0! THAT'S WEIRD FIX IT.")
        L1_from_moon = None
        L2_from_moon = None

    # L45
    theta = np.radians(60)
    L4 = np.squeeze(np.vstack((d * np.cos(theta), d * np.sin(theta), np.zeros_like(theta))).T)
    L5 = np.squeeze(np.vstack((d * np.cos(-theta), d * np.sin(-theta), np.zeros_like(-theta))).T)
    return {
        "L1": L1_from_moon + r,
        "L2": L2_from_moon + r,
        "L3": -r,
        "L4": L4,
        "L5": L5
    }



import pytest
import numpy as np

from astropy.time import Time
import astropy.units as u
from astropy.coordinates import Longitude, Latitude
from astropy.table import QTable

import ssapy
from .ssapy_test_helpers import timer
from ssapy.constants import RGEO, VGEO

def _mock_chain_from_rv(r, v, nSteps=10):
    """ Make a mock MCMC chain output from an orbit model
    """
    params = np.concatenate((r, v))
    chain = np.zeros((nSteps, 6))
    for istep in range(nSteps):
        chain[istep, :] = params
    return chain


def _create_iods(num_epochs=10, num_obs_per_track=3, nBurn=500, nStep=100):
    from ssapy.rvsampler import RPrior, APrior
    np.random.seed(27)

    sigma_arcsec = 1.0
    sigma_rad = sigma_arcsec * np.pi / (180.*3600.)

    time0 = Time(2458316., format='jd')

    # Setup observer location
    lon = 100.0
    lat = 33.0
    elevation = 1300.0
    earthObserver = ssapy.EarthObserver(lon, lat, elevation)

    # Create orbit for target
    orbit = ssapy.Orbit.fromKeplerianElements(RGEO*0.98, 0.01, 0.001, 0.0, 1.2, 1.03, time0)

    # Build observation times
    earthTimes = time0 + np.linspace(0, 4, num_obs_per_track)*u.h
    earthRaDec = ssapy.radec(orbit, earthTimes, observer=earthObserver)
    earthRStation, earthVStation = earthObserver.getRV(earthTimes)

    # Create multiple tracks for a single object
    iods = []
    time = time0
    for iepoch in range(num_epochs):
        print("IOD epoch {:d}".format(iepoch+1))

        arc = QTable()
        err_ra = np.random.randn(len(earthRaDec[0])) * sigma_rad
        err_dec = np.random.randn(len(earthRaDec[1])) * sigma_rad
        arc['ra'] = Longitude((earthRaDec[0] + err_ra)*u.rad)
        arc['dec'] = Latitude((earthRaDec[1] + err_dec)*u.rad)
        arc['rStation_GCRF'] = earthRStation*u.m
        arc['vStation_GCRF'] = earthVStation*u.m/u.s
        arc['time'] = Time(earthTimes)
        arc['sigma'] = sigma_arcsec * np.ones(num_obs_per_track, dtype=float)*u.arcsec

        # initialize somewhere near the truth, but perturbed a bit and with large uncertainties
        r, v = ssapy.rv(orbit, time)
        initializer = ssapy.GaussianRVInitializer(r, v, rSigma=0.1*RGEO, vSigma=0.1*VGEO)
        # Use R and A priors
        priors = [RPrior(RGEO, RGEO*0.2), APrior(RGEO, RGEO*0.2)]

        # Run MCMC sampler to generate particles for the observed epoch
        rvprob = ssapy.RVProbability(arc, time, priors=priors)

        sampler = ssapy.EmceeSampler(rvprob, initializer, nWalker=50)
        chain, lnprob, lnprior = sampler.sample(nBurn=nBurn, nStep=nStep)
        chain, lnprob, lnprior = ssapy.utils.cluster_emcee_walkers(chain, lnprob, lnprior, thresh_multiplier=4)
        chmean = np.mean(chain, axis=(0, 1))
        print("r err:", (chmean[0:3] - r) / r)
        print("v err:", (chmean[3:6] - v) / v)
        print("lnprobs:", np.min(lnprob), np.mean(lnprob), np.max(lnprob))
        #
        # chain = _mock_chain_from_rv(r[0], v[0])

        # write_mcmc_chain(chain, lnprob, "chain_{:d}.pkl".format(iepoch))

        # Create Particles object from sampler output
        iods.append(ssapy.Particles(chain, rvprob, lnpriors=lnprior))

        time += 6.0 * u.h
    return iods


@pytest.mark.slow
@timer
def test_orbit_selector_sampling():
    iods = _create_iods(num_epochs=4, num_obs_per_track=6, nBurn=500, nStep=500)
    lkr = ssapy.Linker(iods)
    print(lkr)

    # for j in range(len(iods)):
    #     for i, iod in enumerate(iods):
    #         print(j,i,iod.lnlike(iods[j].draw_orbit()))

    for itrack in range(lkr.num_tracks):
        lkr.orbit_selectors[itrack] = lkr.sample_orbit_selectors_from_data_conditional(itrack)
    print(lkr)


@pytest.mark.slow
@timer
def test_linker():
    np.random.seed(27)

    time0 = Time(2458316., format='jd')

    num_epochs = 4
    num_obs_per_track = 6

    # iods = _create_iods(num_epochs=num_epochs, num_obs_per_track=6, nBurn=500, nStep=500)
    iods = _create_iods(num_epochs=num_epochs, num_obs_per_track=6, nBurn=500, nStep=100)
    lkr = ssapy.Linker(iods)

    orbit_selector_means = ssapy.ModelSelectorParams(num_epochs, num_epochs, 0.0)
    p_orbit_means = ssapy.ModelSelectorParams(num_epochs, num_epochs, 0.0)

    # Do the linking
    nSteps = 50
    for istep in range(nSteps):
        lkr.update_params_using_carlin_chib()

        # Add to running means
        orbit_selector_means.params += lkr.orbit_selectors.params
        p_orbit_means.params += lkr.p_orbit.params

    orbit_selector_means.params /= float(nSteps)
    p_orbit_means.params /= float(nSteps)

    #
    # Test linking accuracy
    #
    # 1 - Check average values of orbit selector parameters
    print("Mean orbit_selectors:\n", orbit_selector_means)
    # 2 - Check average values of P_orbit parameters
    print("Mean P_orbit:\n", p_orbit_means)


if __name__ == '__main__':
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--prof", action='store_true')
    parser.add_argument("--prof_out", type=str, default=None)
    parser.add_argument("--prof_png", type=str, default=None)
    args = parser.parse_args()

    if args.prof:
        import cProfile
        pr = cProfile.Profile()
        pr.enable()

    test_orbit_selector_sampling()
    test_linker()

    if args.prof:
        import pstats
        pr.disable()
        ps = pstats.Stats(pr).sort_stats('cumtime')
        ps.print_stats(30)
        ps = pstats.Stats(pr).sort_stats('tottime')
        ps.print_stats(30)
        if args.prof_out:
            pr.dump_stats(args.prof_out)
            if args.prof_png:
                import subprocess
                cmd = "gprof2dot -f pstats {} -n1 -e1 | dot -Tpng -o {}".format(args.prof_out, args.prof_png)
                subprocess.run(cmd, shell=True)



import os
import subprocess

def ensure_data_downloaded():
    """
    Lazily clones the ssapy/data directory from GitHub if it doesn't already exist.
    """
    data_dir = os.path.join(os.path.dirname(__file__), 'data')
    if os.path.exists(data_dir) and os.listdir(data_dir):
        # print(f"[ssapy] Data directory already exists at {data_dir}")
        return

    # print("[ssapy] Downloading data directory from GitHub...")

    repo_url = "https://github.com/LLNL/SSAPy.git"
    commit_hash = "0ea3174"
    temp_repo_dir = os.path.join(os.path.dirname(__file__), "_temp_ssapy_repo")

    try:
        subprocess.run(["git", "clone", repo_url, temp_repo_dir], check=True)
        subprocess.run(["git", "-C", temp_repo_dir, "checkout", commit_hash], check=True)

        source_data_dir = os.path.join(temp_repo_dir, "ssapy", "data")

        if not os.path.exists(source_data_dir):
            raise FileNotFoundError(f"Expected data directory not found at {source_data_dir}")

        os.makedirs(data_dir, exist_ok=True)
        subprocess.run(["cp", "-r", source_data_dir + "/.", data_dir], check=True)

        print(f"[ssapy] Data successfully downloaded to {data_dir}")

    except subprocess.CalledProcessError as e:
        # print(f"[ssapy] Error downloading data: {e}")
        raise RuntimeError("Failed to clone and copy data from GitHub.")

    finally:
        subprocess.run(["rm", "-rf", temp_repo_dir], check=True)
